#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from uuid import UUID

from evo import logging
from evo.common import APIConnector, BaseAPIClient, HealthCheckType, ICache, Page, ServiceHealth
from evo.common.data import EmptyResponse, Environment, OrderByOperatorEnum
from evo.common.utils import get_service_health, parse_order_by

from ..data import ObjectMetadata, ObjectOrderByEnum, ObjectReference, ObjectVersion, OrgObjectMetadata, Stage
from ..endpoints import MetadataApi, ObjectsApi, StagesApi
from ..endpoints.models import GeoscienceObject, MetadataUpdateBody, UpdateGeoscienceObject
from ..exceptions import ObjectUUIDError
from ..io import ObjectDataDownload, ObjectDataUpload
from . import parse
from .object_client import DownloadedObject

try:
    from ..utils import ObjectDataClient
except ImportError:
    _DATA_CLIENT_AVAILABLE = False
else:
    _DATA_CLIENT_AVAILABLE = True

logger = logging.getLogger("object.client")

__all__ = ["ObjectAPIClient"]


class ObjectAPIClient(BaseAPIClient):
    def __init__(self, environment: Environment, connector: APIConnector, cache: ICache | None = None) -> None:
        """
        :param environment: The target Evo environment, providing org and workspace IDs.
        :param connector: The API connector to use for making API calls.
        :param cache: An optional cache to use for data downloads.
        """
        super().__init__(environment, connector)
        self._stages_api = StagesApi(connector=connector)
        self._objects_api = ObjectsApi(connector=connector)
        self._metadata_api = MetadataApi(connector=connector)
        self._cache = cache

    async def get_service_health(self, check_type: HealthCheckType = HealthCheckType.FULL) -> ServiceHealth:
        """Get the health of the geoscience object service.

        :param check_type: The type of health check to perform.

        :return: A ServiceHealth object.

        :raises EvoAPIException: If the API returns an unexpected status code.
        :raises ClientValueError: If the response is not a valid service health check response.
        """
        return await get_service_health(self._connector, "geoscience-object", check_type=check_type)

    async def list_objects(
        self,
        offset: int = 0,
        limit: int = 5000,
        order_by: dict[ObjectOrderByEnum | str, OrderByOperatorEnum] | None = None,
        schema_id: list[str] | None = None,
        deleted: bool | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> Page[ObjectMetadata]:
        """List up to `limit` geoscience objects, starting at `offset`.

        The geoscience objects will be the latest version of the object.
        If there are no objects to list, the page will be empty.

        :param offset: The number of objects to skip before listing.
        :param limit: Max number of objects to list.
        :param order_by: A dictionary of fields to order the results by, with the field name as the key and the direction of ordering as the value.
        :param schema_id: A list of schema IDs to filter the objects by. If None, objects of all schema types are returned.
        :param deleted: When true, only objects that have been deleted will be returned. If false or None, only non-deleted objects will be returned.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A page of all objects from the query.
        """
        assert limit > 0
        assert offset >= 0
        parsed_order_by = parse_order_by(order_by)
        response = await self._objects_api.list_objects(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            limit=limit,
            offset=offset,
            order_by=parsed_order_by,
            schema_id=schema_id,
            request_timeout=request_timeout,
            deleted=deleted,
        )
        return parse.page_of_metadata(response, self._environment)

    async def list_all_objects(
        self,
        limit_per_request: int = 5000,
        order_by: dict[ObjectOrderByEnum | str, OrderByOperatorEnum] | None = None,
        schema_id: list[str] | None = None,
        deleted: bool | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> list[ObjectMetadata]:
        """List all geoscience objects in the workspace.

        This method makes multiple calls to the `list_objects` endpoint until all objects have been listed.

        :param limit_per_request: The maximum number of objects to list in one request.
        :param order_by: A dictionary of fields to order the results by, with the field name as the key and the direction of ordering as the value.
        :param schema_id: A list of schema IDs to filter the objects by. If None, objects of all schema types are returned.
        :param deleted: When true, only objects that have been deleted will be returned. If false or None, only non-deleted objects will be returned.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A list of all objects in the workspace.
        """
        items = []
        offset = 0
        while True:
            page = await self.list_objects(
                offset=offset,
                limit=limit_per_request,
                order_by=order_by,
                schema_id=schema_id,
                request_timeout=request_timeout,
                deleted=deleted,
            )
            items += page.items()
            if page.is_last:
                break
            offset = page.next_offset
        return items

    async def list_objects_for_instance(
        self,
        offset: int = 0,
        limit: int = 5000,
        order_by: dict[ObjectOrderByEnum | str, OrderByOperatorEnum] | None = None,
        schema_id: list[str] | None = None,
        deleted: bool | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> Page[OrgObjectMetadata]:
        """List up to `limit` geoscience objects for all accessible workspaces in the instance, starting at `offset`.

        The geoscience objects will be the latest version of the object.
        If there are no objects to list, the page will be empty.

        :param offset: The number of objects to skip before listing.
        :param limit: Max number of objects to list.
        :param order_by: A dictionary of fields to order the results by, with the field name as the key and the direction of ordering as the value.
        :param schema_id: A list of schema IDs to filter the objects by. If None, objects of all schema types are returned.
        :param deleted: When true, only objects that have been deleted will be returned. If false or None, only non-deleted objects will be returned.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A page of all objects from the query.
        """
        assert limit > 0
        assert offset >= 0
        parsed_order_by = parse_order_by(order_by)
        response = await self._objects_api.list_objects_by_org(
            org_id=str(self._environment.org_id),
            limit=limit,
            offset=offset,
            order_by=parsed_order_by,
            schema_id=schema_id,
            request_timeout=request_timeout,
            permitted_workspaces_only=True,
            deleted=deleted,
        )
        return parse.page_of_metadata(response, self._environment)

    async def list_versions_by_path(
        self, path: str, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> list[ObjectVersion]:
        """List all version for the given object.

        :param path: The path to the geoscience object.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A sorted list of object versions. The latest version is the first element of the list.
        """
        response = await self._objects_api.get_object(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            objects_path=path,
            include_versions=True,
            request_timeout=request_timeout,
        )
        return parse.versions(response)

    async def list_versions_by_id(
        self, object_id: UUID, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> list[ObjectVersion]:
        """List all version for the given object.

        :param object_id: The UUID of the geoscience object.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A sorted list of object versions. The latest version is the first element of the list.
        """
        response = await self._objects_api.get_object_by_id(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            object_id=str(object_id),
            include_versions=True,
            request_timeout=request_timeout,
        )
        return parse.versions(response)

    async def prepare_data_upload(self, data_identifiers: Sequence[str | UUID]) -> AsyncIterator[ObjectDataUpload]:
        """Prepare to upload multiple data files to the geoscience object service.

        Referenced data that already exists in the workspace will be skipped.

        :param data_identifiers: A list of sha256 digests or UUIDs for the data to be uploaded.

        :return: An async iterator of data upload contexts that can be used to upload the data. Data identifiers
            that already exist in the workspace will be skipped.
        """
        async for ctx in ObjectDataUpload._create_multiple(
            connector=self._connector,
            environment=self._environment,
            names=data_identifiers,
        ):
            yield ctx

    async def prepare_data_download(
        self,
        object_id: UUID,
        version_id: str,
        data_identifiers: Sequence[str | UUID],
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> AsyncIterator[ObjectDataDownload]:
        """Prepare to download multiple data files from the geoscience object service.

        Any data IDs that are not associated with the requested object will raise a DataNotFoundError.

        :param object_id: The ID of the object to download data from.
        :param version_id: The version ID of the object to download data from.
        :param data_identifiers: A list of sha256 digests or UUIDs for the data to be downloaded.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: An async iterator of data download contexts that can be used to download the data.

        :raises DataNotFoundError: If any requested data ID is not associated with the referenced object.
        """
        downloaded_object = await self.download_object_by_id(
            object_id, version=version_id, request_timeout=request_timeout
        )
        for ctx in downloaded_object.prepare_data_download(data_identifiers):
            yield ctx

    if _DATA_CLIENT_AVAILABLE:
        # Optional data client functionality, enabled if the data client dependencies are installed.

        def get_data_client(self, cache: ICache) -> ObjectDataClient:
            """Get a data client for the geoscience object service.

            The data client provides a high-level interface for uploading and downloading data that is referenced in
            geoscience objects, and caching the data locally. It depends on the optional dependency `pyarrow`, which is
            not installed by default. This dependency can be installed with `pip install evo-objects[utils]`.

            :param cache: The cache to use for data downloads.

            :return: An ObjectDataClient instance.

            :raises RuntimeError: If the `pyarrow` package is not installed.
            """
            return ObjectDataClient(environment=self._environment, connector=self._connector, cache=cache)

    async def create_geoscience_object(
        self, path: str, object_dict: dict, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> ObjectMetadata:
        """Upload a new geoscience object to the geoscience object service.

        New geoscience objects must not have a UUID, so that one can be assigned by the Geoscience Object Service.
        The `object_instance` that is passed in will be updated with the service-assigned UUID after it has been
        uploaded, and the metadata of the created object will be returned.

        :param path: The path to upload the object to.
        :param object_dict: The geoscience object to be uploaded.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The metadata of the uploaded object.

        :raises ObjectUUIDError: If the provided object has a UUID.
        """
        if object_dict.get("uuid") is not None:
            raise ObjectUUIDError("Object has a UUID but new objects should have None")
        object_for_upload = GeoscienceObject.model_validate(object_dict)

        result = await self._objects_api.post_objects(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            objects_path=path,
            geoscience_object=object_for_upload,
            request_timeout=request_timeout,
        )
        object_dict["uuid"] = result.object_id
        return parse.object_metadata(result, self._environment)

    async def move_geoscience_object(
        self, path: str, object_dict: dict, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> ObjectMetadata:
        """Move an existing geoscience object to a new path in the geoscience object service.

        :param path: The new path to move the object to.
        :param object_dict: The geoscience object to be moved.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The metadata of the moved object.

        :raises ObjectUUIDError: If the provided object does not have a UUID.
        """
        if object_dict.get("uuid") is None:
            raise ObjectUUIDError("Object does not have a UUID")
        object_for_upload = GeoscienceObject.model_validate(object_dict)

        result = await self._objects_api.post_objects(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            objects_path=path,
            geoscience_object=object_for_upload,
            request_timeout=request_timeout,
        )
        return parse.object_metadata(result, self._environment)

    async def update_geoscience_object(
        self, object_dict: dict, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> ObjectMetadata:
        """Update an existing geoscience object in the geoscience object service.

        :param object_dict: The geoscience object to be updated.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The metadata of the updated object.

        :raises ObjectUUIDError: If the provided object does not have a UUID.
        """
        if object_dict.get("uuid") is None:
            raise ObjectUUIDError("Object does not have a UUID")
        object_for_upload = UpdateGeoscienceObject.model_validate(object_dict)

        result = await self._objects_api.update_objects_by_id(
            object_id=str(object_for_upload.uuid),
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            update_geoscience_object=object_for_upload,
            request_timeout=request_timeout,
        )
        return parse.object_metadata(result, self._environment)

    async def download_object_by_path(
        self,
        path: str,
        version: str | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> DownloadedObject:
        """Download a geoscience object definition (by path).

        :param path: The path to the geoscience object.
        :param version: The version of the geoscience object to download. This will download the latest version by
            default.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A tuple containing the object metadata and a data model of the requested geoscience object.
        """
        reference = ObjectReference.new(environment=self._environment, object_path=path, version_id=version)
        return await DownloadedObject.from_reference(
            connector=self._connector,
            reference=reference,
            cache=self._cache,
            request_timeout=request_timeout,
        )

    async def download_object_by_id(
        self,
        object_id: UUID,
        version: str | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> DownloadedObject:
        """Download a geoscience object definition (by UUID).

        :param object_id: The uuid of the geoscience object.
        :param version: The version of the geoscience object to download. This will download the latest version by
            default.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A tuple containing the object metadata and a data model of the requested geoscience object.
        """
        reference = ObjectReference.new(environment=self._environment, object_id=object_id, version_id=version)
        return await DownloadedObject.from_reference(
            connector=self._connector,
            reference=reference,
            cache=self._cache,
            request_timeout=request_timeout,
        )

    async def get_latest_object_versions(
        self,
        object_ids: list[UUID],
        batch_size: int = 500,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> dict[UUID, str]:
        """Get the latest version of each object by uuid.

        :param object_ids: A list of object uuids.
        :param batch_size: The maximum number of objects to check in one API call (max 500).
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: A mapping of uuids to the latest version id.
        """
        offset = 0
        n_objects = len(object_ids)
        latest_ids: dict[UUID, str] = {}
        while batch_object_ids := object_ids[offset : min(offset + batch_size, n_objects)]:
            offset += batch_size
            response = await self._objects_api.list_object_version_ids(
                org_id=str(self._environment.org_id),
                workspace_id=str(self._environment.workspace_id),
                request_body=[str(object_id) for object_id in batch_object_ids],
                request_timeout=request_timeout,
            )
            latest_ids.update({UUID(latest.object_id): latest.version_id for latest in response})
        return latest_ids

    async def delete_object_by_path(
        self, path: str, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> None:
        """Soft-delete a geoscience object (by path).

        :param path: The path to the geoscience object.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: None
        """
        await self._objects_api.delete_object_by_path(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            objects_path=path,
            additional_headers={"Accept-Encoding": "gzip"},
            request_timeout=request_timeout,
        )

    async def delete_object_by_id(
        self, object_id: UUID, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> None:
        """Soft-delete a geoscience object (by UUID).

        :param object_id: The uuid of the geoscience object
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: None
        """
        await self._objects_api.delete_objects_by_id(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            object_id=str(object_id),
            additional_headers={"Accept-Encoding": "gzip"},
            request_timeout=request_timeout,
        )

    async def restore_geoscience_object(
        self, object_id: UUID, request_timeout: int | float | tuple[int | float, int | float] | None = None
    ) -> None | ObjectMetadata:
        """Restore a soft-deleted geoscience object in the geoscience object service.

        :param object_id: The uuid of the geoscience object
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The metadata of the restored object, if a rename occurred. Otherwise, None.
        """
        result = await self._objects_api.update_objects_by_id(
            object_id=str(object_id),
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            deleted=False,
            request_timeout=request_timeout,
        )
        # If the restore happened without a rename, the response will be empty
        # If the restore happened with a rename, the response will be the metadata of the restored object
        if isinstance(result, EmptyResponse):
            return None
        return parse.object_metadata(result, self._environment)

    async def list_stages(self) -> list[Stage]:
        """List all available stages in the organisation.

        :return: A list of all available stages."""
        response = await self._stages_api.list_stages(org_id=str(self._environment.org_id))
        return [parse.stage(model) for model in response.stages]

    async def set_stage(self, object_id: UUID, version_id: int, stage_id: UUID) -> None:
        """Set the stage of a specific version of a geoscience object.

        :param object_id: The UUID of the geoscience object.
        :param version_id: The version ID of the geoscience object.
        :param stage_id: The UUID of the stage to set.

        :return: None
        """

        await self._metadata_api.update_metadata(
            object_id=str(object_id),
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            metadata_update_body=MetadataUpdateBody(stage_id=stage_id),
            version_id=version_id,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from unittest import mock
from uuid import UUID

from data import load_test_data
from evo.colormaps import ColormapAPIClient
from evo.colormaps.data import (
    Association,
    AssociationMetadata,
    CategoryColormap,
    ColormapMetadata,
    ContinuousColormap,
    DiscreteColormap,
)
from evo.colormaps.exceptions import UnknownColormapType
from evo.common import Environment, HealthCheckType, RequestMethod, ServiceUser
from evo.common.exceptions import NotFoundException
from evo.common.test_tools import BASE_URL, ORG, WORKSPACE_ID, TestWithConnector, utc_datetime


class TestColormapApiClient(TestWithConnector):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        self.environment = Environment(hub_url=BASE_URL, org_id=ORG.id, workspace_id=WORKSPACE_ID)
        self.colormap_api_client = ColormapAPIClient(connector=self.connector, environment=self.environment)

    @property
    def base_path(self) -> str:
        return f"colormap/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    async def test_check_service_health(self) -> None:
        """Test service health check implementation."""
        with mock.patch("evo.colormaps.client.get_service_health", spec_set=True) as mock_get_service_health:
            await self.colormap_api_client.get_service_health()
        mock_get_service_health.assert_called_once_with(
            self.connector, "visualization", check_type=HealthCheckType.FULL
        )

    async def test_create_continuous_colormap(self) -> None:
        post_colormap_response = load_test_data("continuous_colormap_response.json")
        colormap = ContinuousColormap(
            colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
            attribute_controls=[0.1, 0.2, 0.3],
            gradient_controls=[0.4, 0.5, 0.6],
        )
        colormap_name = "continuous colormap 1"
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=colormap,
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=colormap_name,
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=6)}",
            id=UUID(int=6),
        )
        with self.transport.set_http_response(
            status_code=201, content=json.dumps(post_colormap_response), headers={"Content-Type": "application/json"}
        ):
            colormap_metadata = await self.colormap_api_client.create_colormap(colormap=colormap, name=colormap_name)

        self.assertIsInstance(colormap_metadata, ColormapMetadata)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/colormaps",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "attribute_controls": colormap.attribute_controls,
                "colors": colormap.colors,
                "dtype": "continuous",
                "gradient_controls": colormap.gradient_controls,
                "name": colormap_name,
            },
        )

    async def test_create_discrete_colormap(self) -> None:
        post_colormap_response = load_test_data("discrete_colormap_response.json")
        colormap = DiscreteColormap(
            colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
            end_inclusive=[True, False],
            end_points=[0.1, 0.2],
        )
        colormap_name = "discrete colormap 1"
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=colormap,
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=colormap_name,
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=7)}",
            id=UUID(int=7),
        )
        with self.transport.set_http_response(
            status_code=201, content=json.dumps(post_colormap_response), headers={"Content-Type": "application/json"}
        ):
            colormap_metadata = await self.colormap_api_client.create_colormap(colormap=colormap, name=colormap_name)

        self.assertIsInstance(colormap_metadata, ColormapMetadata)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/colormaps",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "colors": colormap.colors,
                "dtype": "discrete",
                "end_inclusive": colormap.end_inclusive,
                "end_points": colormap.end_points,
                "name": colormap_name,
            },
        )

    async def test_create_category_colormap(self) -> None:
        post_colormap_response = load_test_data("category_colormap_response.json")
        colormap = CategoryColormap(
            colors=[[0, 0, 219], [0, 217, 0], [217, 0, 0]],
            map=["a", "b", "c"],
        )
        colormap_name = "category colormap 1"
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=colormap,
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=colormap_name,
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=8)}",
            id=UUID(int=8),
        )
        with self.transport.set_http_response(
            status_code=201, content=json.dumps(post_colormap_response), headers={"Content-Type": "application/json"}
        ):
            colormap_metadata = await self.colormap_api_client.create_colormap(colormap=colormap, name=colormap_name)

        self.assertIsInstance(colormap_metadata, ColormapMetadata)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/colormaps",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "colors": colormap.colors,
                "dtype": "category",
                "map": colormap.map,
                "name": colormap_name,
            },
        )

    async def test_create_duplicate_colourmap(self) -> None:
        post_colormap_response = load_test_data("continuous_colormap_response.json")
        colormap = ContinuousColormap(
            colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
            attribute_controls=[0.1, 0.2, 0.3],
            gradient_controls=[0.4, 0.5, 0.6],
        )
        colormap_name = "continuous colormap 1"
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=colormap,
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=colormap_name,
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=6)}",
            id=UUID(int=6),
        )
        with self.transport.set_http_response(
            status_code=201, content=json.dumps(post_colormap_response), headers={"Content-Type": "application/json"}
        ):
            colormap_metadata = await self.colormap_api_client.create_colormap(colormap=colormap, name=colormap_name)

        self.assertIsInstance(colormap_metadata, ColormapMetadata)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/colormaps",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "attribute_controls": colormap.attribute_controls,
                "colors": colormap.colors,
                "dtype": "continuous",
                "gradient_controls": colormap.gradient_controls,
                "name": colormap_name,
            },
        )

    async def test_create_colormap_unknown_type(self) -> None:
        colormap = "not a colormap"
        with self.assertRaises(UnknownColormapType):
            await self.colormap_api_client.create_colormap(colormap=colormap, name="colormap name")

    async def assert_get_colormap_by_id(self, colormap_id, expected_colormap_metadata, get_colormap_response):
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_colormap_response), headers={"Content-Type": "application/json"}
        ):
            colormap_metadata = await self.colormap_api_client.get_colormap_by_id(colormap_id)
        self.assertIsInstance(colormap_metadata, ColormapMetadata)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/colormaps/{colormap_id}",
            headers={"Accept": "application/json"},
        )

    async def test_get_colormap_by_id_continuous(self) -> None:
        get_colormap_response = load_test_data("continuous_colormap_response.json")
        colormap_id = UUID(int=6)
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=ContinuousColormap(
                colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
                attribute_controls=[0.1, 0.2, 0.3],
                gradient_controls=[0.4, 0.5, 0.6],
            ),
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name="continuous colormap 1",
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            id=colormap_id,
        )
        await self.assert_get_colormap_by_id(colormap_id, expected_colormap_metadata, get_colormap_response)

    async def test_get_colormap_by_id_discrete(self) -> None:
        get_colormap_response = load_test_data("discrete_colormap_response.json")
        colormap_id = UUID(int=7)
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=DiscreteColormap(
                colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
                end_inclusive=[True, False],
                end_points=[0.1, 0.2],
            ),
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name="discrete colormap 1",
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            id=colormap_id,
        )
        await self.assert_get_colormap_by_id(colormap_id, expected_colormap_metadata, get_colormap_response)

    async def test_get_colormap_by_id_category(self) -> None:
        get_colormap_response = load_test_data("category_colormap_response.json")
        colormap_id = UUID(int=8)
        expected_colormap_metadata = ColormapMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            colormap=CategoryColormap(
                colors=[[0, 0, 219], [0, 217, 0], [217, 0, 0]],
                map=["a", "b", "c"],
            ),
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name="category colormap 1",
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            id=colormap_id,
        )
        await self.assert_get_colormap_by_id(colormap_id, expected_colormap_metadata, get_colormap_response)

    async def test_get_colormap_by_id_does_not_exist(self) -> None:
        colormap_id = UUID(int=9)
        with self.transport.set_http_response(status_code=404, content=""):
            with self.assertRaises(NotFoundException):
                await self.colormap_api_client.get_colormap_by_id(colormap_id)

    async def test_get_colormap_collection(self) -> None:
        get_colormap_collection_response = load_test_data("colormap_collection_response.json")
        expected_colormap_metadata = [
            ColormapMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                colormap=ContinuousColormap(
                    colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
                    attribute_controls=[0.1, 0.2, 0.3],
                    gradient_controls=[0.4, 0.5, 0.6],
                ),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name="continuous colormap 1",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=6)}",
                id=UUID(int=6),
            ),
            ColormapMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                colormap=DiscreteColormap(
                    colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
                    end_inclusive=[True, False],
                    end_points=[0.1, 0.2],
                ),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name="discrete colormap 1",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=7)}",
                id=UUID(int=7),
            ),
            ColormapMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                colormap=CategoryColormap(
                    colors=[[255, 240, 219], [238, 217, 196], [217, 185, 155]],
                    map=["a", "b", "c"],
                ),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name="category colormap 1",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{UUID(int=8)}",
                id=UUID(int=8),
            ),
        ]
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(get_colormap_collection_response),
            headers={"Content-Type": "application/json"},
        ):
            colormap_metadata = await self.colormap_api_client.get_colormap_collection()

        self.assertIsInstance(colormap_metadata, list)
        self.assertEqual(colormap_metadata, expected_colormap_metadata)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/colormaps",
            headers={"Accept": "application/json"},
        )

    async def test_create_association(self) -> None:
        colormap_association_response = load_test_data("colormap_association_response.json")
        colormap_id = UUID(int=6)
        attribute_id = "a very unique ID"
        association = Association(attribute_id=attribute_id, colormap_id=colormap_id)
        object_id = UUID(int=20)
        association_id = UUID(int=30)

        association_metadata = AssociationMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=f"Association between {attribute_id} and {colormap_id}",
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects/{object_id}/associations/{association_id}",
            id=association_id,
            attribute_id=attribute_id,
            object_id=object_id,
            colormap_id=colormap_id,
            colormap_uri=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
        )
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(colormap_association_response),
            headers={"Content-Type": "application/json"},
        ):
            association = await self.colormap_api_client.create_association(object_id, association)

        self.assertIsInstance(association, AssociationMetadata)
        self.assertEqual(association, association_metadata)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={"attribute_id": str(attribute_id), "colormap_id": str(colormap_id)},
        )

    async def test_create_batch_associations(self) -> None:
        colormap_association_response = load_test_data("colormap_association_collection_response.json")
        object_id = UUID(int=20)
        association_ids = [UUID(int=30), UUID(int=31), UUID(int=32)]
        attribute_ids = ["a fairly unique ID", str(UUID(int=2)), "another fairly unique ID"]
        colormap_ids = [UUID(int=6), UUID(int=7), UUID(int=8)]
        associations = [
            Association(attribute_id=attribute_id, colormap_id=colormap_id)
            for attribute_id, colormap_id in zip(attribute_ids, colormap_ids)
        ]

        expected_associations = [
            AssociationMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name=f"Association between {attribute_id} and {colormap_id}",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects/{object_id}/associations/{association_id}",
                id=association_id,
                colormap_id=colormap_id,
                attribute_id=attribute_id,
                object_id=object_id,
                colormap_uri=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            )
            for association_id, attribute_id, colormap_id in zip(association_ids, attribute_ids, colormap_ids)
        ]
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(colormap_association_response),
            headers={"Content-Type": "application/json"},
        ):
            associations = await self.colormap_api_client.create_batch_associations(object_id, associations)

        self.assertIsInstance(associations, list)
        self.assertEqual(associations, expected_associations)
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations/batch",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "associations": [
                    {"attribute_id": attribute_id, "colormap_id": str(colormap_id)}
                    for attribute_id, colormap_id in zip(attribute_ids, colormap_ids)
                ]
            },
        )

    async def test_create_batch_associations_chunked(self) -> None:
        colormap_association_response = load_test_data("colormap_association_collection_response.json")
        object_id = UUID(int=20)
        associations = [Association(attribute_id=f"attribute_id_{i}", colormap_id=UUID(int=i)) for i in range(500)]
        association_ids = [UUID(int=30), UUID(int=31), UUID(int=32)]
        attribute_ids = ["a fairly unique ID", str(UUID(int=2)), "another fairly unique ID"]
        colormap_ids = [UUID(int=6), UUID(int=7), UUID(int=8)]
        expected_associations = [
            AssociationMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name=f"Association between {attribute_id} and {colormap_id}",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects/{object_id}/associations/{association_id}",
                id=association_id,
                colormap_id=colormap_id,
                attribute_id=attribute_id,
                object_id=object_id,
                colormap_uri=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            )
            for association_id, attribute_id, colormap_id in zip(association_ids, attribute_ids, colormap_ids)
        ] * 4  # The requests should be chunked 4 times and stitch the results back together
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(colormap_association_response),
            headers={"Content-Type": "application/json"},
        ):
            associations = await self.colormap_api_client.create_batch_associations(object_id, associations)

        self.assertIsInstance(associations, list)
        self.assertEqual(associations, expected_associations)
        self.assert_any_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations/batch",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "associations": [
                    {"attribute_id": f"attribute_id_{i}", "colormap_id": str(UUID(int=i))} for i in range(128)
                ]
            },
        )
        self.assert_any_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations/batch",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "associations": [
                    {"attribute_id": f"attribute_id_{i}", "colormap_id": str(UUID(int=i))} for i in range(128, 256)
                ]
            },
        )
        self.assert_any_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations/batch",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "associations": [
                    {"attribute_id": f"attribute_id_{i}", "colormap_id": str(UUID(int=i))} for i in range(256, 384)
                ]
            },
        )
        self.assert_any_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{object_id}/associations/batch",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body={
                "associations": [
                    {"attribute_id": f"attribute_id_{i}", "colormap_id": str(UUID(int=i))} for i in range(384, 500)
                ]
            },
        )

    async def test_create_batch_associations_empty(self) -> None:
        object_id = UUID(int=20)
        associations = []
        with self.assertRaises(ValueError):
            await self.colormap_api_client.create_batch_associations(object_id, associations)

    async def test_get_association(self) -> None:
        colormap_association_response = load_test_data("colormap_association_response.json")
        colormap_id = UUID(int=6)
        attribute_id = "a very unique ID"
        object_id = UUID(int=20)
        association_id = UUID(int=30)

        expected_association_metadata = AssociationMetadata(
            environment=self.environment,
            created_at=utc_datetime(2024, 9, 16, 1, 30),
            created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            modified_at=utc_datetime(2024, 9, 16, 1, 30),
            modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
            name=f"Association between {attribute_id} and {colormap_id}",
            self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects/{object_id}/associations/{association_id}",
            id=association_id,
            colormap_id=colormap_id,
            attribute_id=attribute_id,
            object_id=object_id,
            colormap_uri=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
        )
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(colormap_association_response),
            headers={"Content-Type": "application/json"},
        ):
            association = await self.colormap_api_client.get_association(object_id, association_id)

        self.assertIsInstance(association, AssociationMetadata)
        self.assertEqual(association, expected_association_metadata)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}/associations/{association_id}",
            headers={"Accept": "application/json"},
        )

    async def test_get_association_collection(self) -> None:
        colormap_association_response = load_test_data("colormap_association_collection_response.json")
        object_id = UUID(int=20)
        association_ids = [UUID(int=30), UUID(int=31), UUID(int=32)]
        attribute_ids = ["a fairly unique ID", str(UUID(int=2)), "another fairly unique ID"]
        colormap_ids = [UUID(int=6), UUID(int=7), UUID(int=8)]

        expected_associations = [
            AssociationMetadata(
                environment=self.environment,
                created_at=utc_datetime(2024, 9, 16, 1, 30),
                created_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                modified_at=utc_datetime(2024, 9, 16, 1, 30),
                modified_by=ServiceUser(id=UUID(int=16), name=None, email=None),
                name=f"Association between {attribute_id} and {colormap_id}",
                self_link=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects/{object_id}/associations/{association_id}",
                id=association_id,
                colormap_id=colormap_id,
                attribute_id=attribute_id,
                object_id=object_id,
                colormap_uri=f"{BASE_URL}colormap/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/colormaps/{colormap_id}",
            )
            for association_id, attribute_id, colormap_id in zip(association_ids, attribute_ids, colormap_ids)
        ]
        with self.transport.set_http_response(
            status_code=200,
            content=json.dumps(colormap_association_response),
            headers={"Content-Type": "application/json"},
        ):
            associations = await self.colormap_api_client.get_association_collection(object_id)

        self.assertIsInstance(associations, list)
        self.assertEqual(associations, expected_associations)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{object_id}/associations",
            headers={"Accept": "application/json"},
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# generated by datamodel-codegen:
#   filename:  spec.yaml

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Literal
from uuid import UUID

from pydantic import AnyUrl, Field, StrictBool, StrictFloat, StrictInt, StrictStr

from .._model_config import CustomBaseModel


class BlockSize(CustomBaseModel):
    x: StrictFloat = Field(..., title="X")
    y: StrictFloat = Field(..., title="Y")
    z: StrictFloat = Field(..., title="Z")


class ColumnHeaderType(Enum):
    id = "id"
    name = "name"


class ColumnNameMapping(CustomBaseModel):
    csv_name: StrictStr = Field(..., title="CSV Name")
    new_title: StrictStr = Field(..., title="New Title")


class ColumnRename(CustomBaseModel):
    new_title: StrictStr = Field(..., title="New Title")
    old_title: StrictStr = Field(..., title="Old Title")


class DataType(Enum):
    """
       Capitalized list of supported PyArrow datatypes outlined in
       [https://arrow.apache.org/docs/python/api/datatypes.html](https://arrow.apache.org/docs/python/api/datatypes.html).

    Timestamp datatype is expected to be used with unit measured in "us" and timezone set to "UTC" inside of parquet
        column header schema or it won't be accepted; an example for PyArrow binding would look like this:
        pa.timestamp("us", tz="UTC")

    """

    Boolean = "Boolean"
    Int8 = "Int8"
    Int16 = "Int16"
    Int32 = "Int32"
    Int64 = "Int64"
    UInt8 = "UInt8"
    UInt16 = "UInt16"
    UInt32 = "UInt32"
    UInt64 = "UInt64"
    Float16 = "Float16"
    Float32 = "Float32"
    Float64 = "Float64"
    Utf8 = "Utf8"
    Date32 = "Date32"
    Timestamp = "Timestamp"


class FloatRange(CustomBaseModel):
    max: StrictFloat = Field(..., title="Max")
    """
    Maximum value for the column
    """
    min: StrictFloat = Field(..., title="Min")
    """
    Minimum value for the column
    """


class GeometryColumns(Enum):
    indices = "indices"
    coordinates = "coordinates"


class InputOptionsCSV(CustomBaseModel):
    column_name_mapping: list[ColumnNameMapping] | None = Field([], title="Column Name Mapping")
    decimal_char: StrictStr | None = Field(".", max_length=1, min_length=1, title="Decimal Char")
    delimiter: StrictStr | None = Field(",", max_length=1, min_length=1, title="Delimiter")
    file_format: Literal["csv"] = Field("csv", title="File Format")
    quote_char: StrictStr | None = Field('"', max_length=1, min_length=1, title="Quote Char")
    skip_rows: StrictInt | None = Field(0, ge=0, title="Skip Rows")
    skip_rows_after_headers: StrictInt | None = Field(0, ge=0, title="Skip Rows After Headers")


class InputOptionsParquet(CustomBaseModel):
    file_format: Literal["parquet"] = Field("parquet", title="File Format")


class IntRange(CustomBaseModel):
    """
    Represents a range of whole numbers from min to max inclusive.
    """

    max: StrictInt = Field(..., title="Max")
    """
    Maximum value for the column
    """
    min: StrictInt = Field(..., title="Min")
    """
    Minimum value for the column
    """


class JobErrorPayload(CustomBaseModel):
    detail: StrictStr = Field(
        ...,
        examples=["Update file 38e3ca486b3b477db0f99646b0529746 not found."],
        title="Detail",
    )
    """
    A human-readable explanation specific to this occurrence of the problem.
    """
    download_url: AnyUrl | None = Field(None, title="Download Url")
    """
    URL for downloading an errors block model file. Set if the job is for block model update and there were block-specific errors.
    """
    status: StrictInt = Field(..., examples=[400], title="Status")
    """
    Status code for the error, equivalent to a HTTP status code.
    """
    title: StrictStr = Field(..., examples=["Error in Update"], title="Title")
    """
    A short, human-readable summary of the problem type.
    """
    type: StrictStr = Field(
        ...,
        examples=["https://seequent.com/error-codes/block-model-service/job/update-error"],
        title="Type",
    )
    """
    A URI reference that is the primary identifier of the problem type.
    """


class JobPendingPayload(CustomBaseModel):
    upload_url: AnyUrl | None = Field(None, title="Upload Url")
    """
    Upload URL for the job, with an up-to-date SAS token
    """


class JobStatus(Enum):
    PENDING_UPLOAD = "PENDING_UPLOAD"
    QUEUED = "QUEUED"
    PROCESSING = "PROCESSING"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"


class Location(CustomBaseModel):
    x: StrictFloat = Field(..., title="X")
    y: StrictFloat = Field(..., title="Y")
    z: StrictFloat = Field(..., title="Z")


class OctreeSubblocks(CustomBaseModel):
    nx: StrictInt = Field(..., title="Nx")
    ny: StrictInt = Field(..., title="Ny")
    nz: StrictInt = Field(..., title="Nz")


class OutputOptionsCSV(CustomBaseModel):
    column_headers: ColumnHeaderType | None = "name"
    """
    Header values within the output file for non-system columns, either the column's name or UUID
    """
    delimiter: StrictStr | None = Field(",", max_length=1, min_length=1, title="Delimiter")
    """
    Delimiter between values
    """
    exclude_null_rows: StrictBool | None = Field(True, title="Exclude Null Rows")
    """
    Specifies whether or not the output query file should exclude rows where all values for queried user columns are null.
    """
    file_format: Literal["csv"] = Field("csv", title="File Format")
    """
    CSV file
    """


class OutputOptionsParquet(CustomBaseModel):
    column_headers: ColumnHeaderType | None = "name"
    """
    Header values within the output file for non-system columns, either the column's name or UUID
    """
    exclude_null_rows: StrictBool | None = Field(True, title="Exclude Null Rows")
    """
    Specifies whether or not the output query file should exclude rows where all values for queried user columns are null.
    """
    file_format: Literal["parquet"] = Field("parquet", title="File Format")
    """
    Apache Parquet file
    """


class QueryDownload(CustomBaseModel):
    download_url: AnyUrl = Field(
        ...,
        examples=[
            "https://au.static.evo.seequent.com/blocksync/169ccde2-4d0f-4a7a-adb7-1ccadc734cdd/query/{criteria hash}.parquet?st=2023-11-09T03%3A05%3A30Z&se=2023-11-10T03%3A06%3A30Z&sp=r&spr=http&sv=2021-08-06&sr=b&sig=ABCDEF0123456789ABCDEF0123456789"
        ],
        title="Download Url",
    )
    """
    Download URL for the results of the query.
    """


class RegularSubblocks(CustomBaseModel):
    nx: StrictInt = Field(..., title="Nx")
    ny: StrictInt = Field(..., title="Ny")
    nz: StrictInt = Field(..., title="Nz")


class Rename(CustomBaseModel):
    col_id: StrictStr = Field(..., examples=["618d6339-2fa7-4dfd-9c7f-c0b12016639e"], title="Col Id")
    """
    The ID of the column, a UUID for non-system columns
    """
    new_title: StrictStr = Field(..., title="New Title")
    """
    The new human-readable label for the column identified by `col_id`
    """


class RenameLite(CustomBaseModel):
    new_title: StrictStr = Field(..., title="New Title")
    """
    The new human-readable label for the column identified by `title`
    """
    title: StrictStr = Field(..., title="Title")
    """
    The original human-readable label used to identify the column
    """


class ReportAggregation(Enum):
    SUM = "SUM"
    MASS_AVERAGE = "MASS_AVERAGE"


class ReportCategory(CustomBaseModel):
    col_id: UUID = Field(..., examples=["13c277c2-edc6-4a7a-8380-251dd19231f2"], title="Col Id")
    """
    ID of the category column in the block model
    """
    label: StrictStr = Field(..., title="Label")
    """
    The human-readable label used to identify the column
    """
    values: list[StrictStr] | None = Field(
        None,
        examples=[["low", "medium", "high"]],
        max_length=100,
        min_length=1,
        title="Values",
    )
    """
    Values to use for category filtering. If null the report will contain a row for each distinct non-empty value in this column.
    """


class ReportColumn(CustomBaseModel):
    aggregation: ReportAggregation = Field(..., examples=["SUM"])
    """
    The aggregation method to use for the column
    """
    col_id: UUID = Field(..., examples=["13c277c2-edc6-4a7a-8380-251dd19231f2"], title="Col Id")
    """
    ID of the column in the block model
    """
    label: StrictStr = Field(..., examples=["Au Content"], title="Label")
    """
    The human-readable label used to identify the column
    """
    output_unit_id: StrictStr = Field(..., examples=["kg"], title="Output Unit Id")
    """
    ID of the unit to use for the column output
    """


class ReportComparisonJobResult(CustomBaseModel):
    from_result_uuid: UUID | None = Field(
        None,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="From Result Uuid",
    )
    """
    ID of the Report Result for the baseline version.
    """
    from_version_uuid: UUID | None = Field(
        None,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="From Version Uuid",
    )
    """
    ID of the block model version to use as the baseline for the comparison.
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the Report Specification that the comparison is for
    """
    to_result_uuid: UUID | None = Field(None, examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="To Result Uuid")
    """
    ID of the Report Result for the version to compare.
    """
    to_version_uuid: UUID | None = Field(
        None, examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="To Version Uuid"
    )
    """
    ID of the block model version to compare.
    """


class ReportComparisonRequestResult(CustomBaseModel):
    from_result_uuid: UUID | None = Field(
        None,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="From Result Uuid",
    )
    """
    ID of the Report Result for the baseline version, or null if the result does not yet exist (and a job was created).
    """
    job_url: AnyUrl | None = Field(None, title="Job Url")
    """
    URL to the Report Comparison Job, if created. If both results already exist, no job will be created and this will be null.
    """
    to_result_uuid: UUID | None = Field(None, examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="To Result Uuid")
    """
    ID of the Report Result for the version to compare, or null if the result does not yet exist (and a job was created).
    """


class ReportComparisonSpec(CustomBaseModel):
    from_version_uuid: UUID | None = Field(
        None,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="From Version Uuid",
    )
    """
    ID of the block model version to use as the baseline for the comparison.
    """
    to_version_uuid: UUID | None = Field(
        None, examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="To Version Uuid"
    )
    """
    ID of the block model version to compare.
    """


class ReportComparisonValue(CustomBaseModel):
    difference: StrictFloat | StrictInt = Field(..., title="Difference")
    from_value: StrictFloat | StrictInt = Field(..., title="From Value")
    percent: StrictFloat | None = Field(None, title="Percent")
    to_value: StrictFloat | StrictInt = Field(..., title="To Value")


class ReportNegativeValuesPolicy(Enum):
    USE = "USE"
    ZERO = "ZERO"
    IGNORE_BLOCK = "IGNORE_BLOCK"


class ReportNullValuesPolicy(Enum):
    ZERO = "ZERO"
    IGNORE_BLOCK = "IGNORE_BLOCK"


class ReportResultCategory(CustomBaseModel):
    col_id: UUID = Field(..., examples=["ddd277c2-edc6-4a7a-8380-251dd19231f2"], title="Col Id")
    """
    ID of the block model column
    """
    label: StrictStr = Field(..., examples=["Rock type"], title="Label")
    """
    Label of the column
    """


class ReportResultColumn(CustomBaseModel):
    col_id: UUID | None = Field(None, examples=["eee277c2-edc6-4a7a-8380-251dd19231f2"], title="Col Id")
    """
    ID of the block model column, if applicable. This will be null for system-calculated columns such as 'Mass'
    """
    label: StrictStr = Field(..., examples=["Average gold content"], title="Label")
    """
    Label of the column
    """
    unit_id: StrictStr = Field(..., examples=["ppm[mass]"], title="Unit Id")
    """
    ID of the unit of the column
    """


class ReportResultReferencedColumn(CustomBaseModel):
    col_id: UUID = Field(..., examples=["eee277c2-edc6-4a7a-8380-251dd19231f2"], title="Col Id")
    """
    ID of the block model column
    """
    data_type: StrictStr = Field(..., examples=["Utf8"], title="Data Type")
    """
    Data type of the referenced column in the Report Result
    """
    title: StrictStr = Field(..., examples=["Something"], title="Title")
    """
    Title of the referenced column
    """
    unit_id: StrictStr | None = Field(None, examples=["ppm[mass]"], title="Unit Id")
    """
    ID of the unit of the column
    """
    unit_id_from_version: UUID | None = Field(
        None,
        examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Unit Id From Version",
    )
    """
    Version UUID of the version of block model that the unid was from
    """


class ReportRow(CustomBaseModel):
    categories: list[StrictStr | StrictInt | StrictFloat | StrictBool | None] = Field(
        ..., examples=[["Granite"]], title="Categories"
    )
    """
    List of category values. `null` indicates a total for that category column.
    """
    values: list[StrictInt | StrictFloat] = Field(..., examples=[[2.7, 0.3]], title="Values")
    """
    List of values for the value columns
    """


class ReportRunResult(CustomBaseModel):
    bm_uuid: UUID = Field(..., title="Bm Uuid")
    report_result_uuid: UUID = Field(..., title="Report Result Uuid")
    report_specification_uuid: UUID = Field(..., title="Report Specification Uuid")
    version_id: StrictInt = Field(..., title="Version Id")
    version_uuid: UUID = Field(..., title="Version Uuid")


class ReportWarningType(Enum):
    INVALID_DENSITY_VALUE = "INVALID_DENSITY_VALUE"
    NULL_CUTOFF_VALUE = "NULL_CUTOFF_VALUE"
    NEGATIVE_CUTOFF_VALUE = "NEGATIVE_CUTOFF_VALUE"
    NULL_COLUMN_VALUE = "NULL_COLUMN_VALUE"
    NEGATIVE_COLUMN_VALUE = "NEGATIVE_COLUMN_VALUE"
    INVALID_CATEGORY_VALUE = "INVALID_CATEGORY_VALUE"
    TOO_MANY_CATEGORY_VALUES = "TOO_MANY_CATEGORY_VALUES"
    LONG_CATEGORY_VALUE = "LONG_CATEGORY_VALUE"
    UNIT_FROM_LATEST_VERSION = "UNIT_FROM_LATEST_VERSION"
    MISSING_SUB_BLOCK = "MISSING_SUB_BLOCK"


class ReportingJobResult(CustomBaseModel):
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model
    """
    job_url: AnyUrl = Field(..., title="Job Url")
    job_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Job Uuid")
    """
    ID of the Reporting Job.
    """
    report_specification_uuid: UUID = Field(..., title="Report Specification Uuid")
    version_id: StrictInt = Field(..., examples=[5], title="Version Id")
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """
    version_uuid: UUID = Field(..., examples=["3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7"], title="Version Uuid")
    """
    A universally unique identifier for the version
    """


class ReportingJobSpec(CustomBaseModel):
    version_uuid: UUID | None = Field(None, examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Version Uuid")
    """
    ID of the block model version to run the report against. If not provided, the latest will be used.
    """


class RotationAxis(Enum):
    x = "x"
    y = "y"
    z = "z"


class Size3D(CustomBaseModel):
    nx: StrictInt = Field(..., title="Nx")
    ny: StrictInt = Field(..., title="Ny")
    nz: StrictInt = Field(..., title="Nz")


class SizeOptionsFlexible(CustomBaseModel):
    model_type: Literal["flexible"] = Field(title="Model Type")
    """
    Type of sub-blocking
    """
    n_parent_blocks: Size3D
    """
    Number of parent blocks for x, y and z axis. Total must not exceed 500,000,000.
    """
    n_subblocks_per_parent: RegularSubblocks
    """
    Number of sub-blocks per parent block per axis.
    """
    parent_block_size: BlockSize
    """
    Parent block size.
    """


class SizeOptionsFullySubBlocked(CustomBaseModel):
    model_type: Literal["fully-sub-blocked"] = Field(title="Model Type")
    """
    Type of sub-blocking
    """
    n_parent_blocks: Size3D
    """
    Number of parent blocks for x, y and z axis. Total must not exceed 500,000,000.
    """
    n_subblocks_per_parent: RegularSubblocks
    """
    Number of sub-blocks per parent block per axis.
    """
    parent_block_size: BlockSize
    """
    Parent block size.
    """


class SizeOptionsOctree(CustomBaseModel):
    model_type: Literal["variable-octree"] = Field(title="Model Type")
    """
    Type of sub-blocking
    """
    n_parent_blocks: Size3D
    """
    Number of parent blocks for x, y and z axis. Total must not exceed 500,000,000.
    """
    n_subblocks_per_parent: OctreeSubblocks
    """
    Number of sub-blocks per parent block per axis. Accepted values per axis are 1, 2, 4, 8, 16, 32 or 64.
    """
    parent_block_size: BlockSize
    """
    Parent block size.
    """


class SizeOptionsRegular(CustomBaseModel):
    block_size: BlockSize
    """
    Block size
    """
    model_type: Literal["regular"] = Field(title="Model Type")
    """
    Type of sub-blocking
    """
    n_blocks: Size3D
    """
    Number of blocks for x, y and z axis. Total must not exceed 500,000,000.
    """


class UnitType(Enum):
    LENGTH = "LENGTH"
    MASS = "MASS"
    VOLUME = "VOLUME"
    VALUE = "VALUE"
    MASS_PER_VOLUME = "MASS_PER_VOLUME"
    MASS_PER_MASS = "MASS_PER_MASS"
    VOLUME_PER_VOLUME = "VOLUME_PER_VOLUME"
    VALUE_PER_MASS = "VALUE_PER_MASS"


class UpdateBlockModel(CustomBaseModel):
    coordinate_reference_system: StrictStr | None = Field(
        None,
        examples=["EPSG:3395"],
        max_length=10000,
        title="Coordinate Reference System",
    )
    """

    Coordinate reference system used in the block model.

    This may be an EPSG code, signified by the prefix 'EPSG:', or a coordinate system definition in WKT format.

    """
    description: StrictStr | None = Field(None, max_length=500, title="Description")
    """
    Block model description
    """
    name: StrictStr | None = Field(None, max_length=100, min_length=3, title="Name")
    r"""
    Name of the block model. This may not contain `/` nor `\`.
    """
    size_unit_id: StrictStr | None = Field(None, title="Size Unit ID")
    """
    Unit ID denoting the length unit used for the block model's blocks.
    """


class UpdateMetadataValues(CustomBaseModel):
    title: StrictStr | None = Field(None, title="Title")
    """
    The new human-readable label for the column
    """
    unit_id: StrictStr | None = Field(None, title="Unit Id")
    """
    The new unit ID for the column
    """


class UpdateType(Enum):
    merge = "merge"
    replace = "replace"


class UpdatedResponse(CustomBaseModel):
    job_url: AnyUrl = Field(..., title="Job Url")


class UserInfo(CustomBaseModel):
    email: StrictStr | None = Field(None, examples=["kim@example.test"], title="Email Address")
    """
    The primary email address of the user. Can be null if an error occurred while retrieving this information.
    """
    id: UUID = Field(..., examples=["59b73891-5538-4e45-ae67-f8c5b00d7405"], title="User ID")
    """
    The ID of the user
    """
    name: StrictStr | None = Field(None, examples=["Kim Paul"], title="Full Name")
    """
    The full name of the user. Can be null if an error occurred while retrieving this information.
    """


class BBox(CustomBaseModel):
    i_minmax: IntRange
    """
    Minimum and maximum values for the `i` column in the file
    """
    j_minmax: IntRange
    """
    Minimum and maximum values for the `j` column in the file
    """
    k_minmax: IntRange
    """
    Minimum and maximum values for the `k` column in the file
    """


class BBoxXYZ(CustomBaseModel):
    x_minmax: FloatRange
    """
    Minimum and maximum values for the `x` column in the file
    """
    y_minmax: FloatRange
    """
    Minimum and maximum values for the `y` column in the file
    """
    z_minmax: FloatRange
    """
    Minimum and maximum values for the `z` column in the file
    """


class Column(CustomBaseModel):
    col_id: StrictStr = Field(..., examples=["618d6339-2fa7-4dfd-9c7f-c0b12016639e"], title="Col Id")
    """
    The ID of the column, a UUID for non-system columns
    """
    data_type: DataType
    title: StrictStr = Field(..., title="Title")
    """
    The human-readable label used to identify the column
    """
    unit_id: StrictStr | None = Field(None, title="Unit Id")
    """
    The ID of the column's unit
    """


class ColumnLite(CustomBaseModel):
    data_type: DataType
    title: StrictStr = Field(..., title="Title")
    """
    The human-readable label used to identify the column
    """
    unit_id: StrictStr | None = Field(None, title="Unit Id")
    """
    The ID of the column's unit
    """


class CreateReportSpecification(CustomBaseModel):
    autorun: StrictBool = Field(True, title="Autorun")
    """
    Whether to automatically run this Report Specification when a new version is published
    """
    bbox: BBox | BBoxXYZ | None = Field(
        None,
        examples=[
            {
                "i_minmax": {"max": 1, "min": 0},
                "j_minmax": {"max": 1, "min": 0},
                "k_minmax": {"max": 1, "min": 0},
            }
        ],
        title="Bbox",
    )
    """
    Bounding box for the report
    """
    categories: list[ReportCategory] | None = Field(
        None,
        examples=[
            [
                {
                    "col_id": "11c277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Grade",
                    "values": ["low", "medium", "high"],
                }
            ]
        ],
        max_length=1,
        min_length=0,
        title="Categories",
    )
    """
    Category columns within this Report Specification. If null or empty, the report will have a single total row for each cut-off.
    """
    columns: list[ReportColumn] = Field(
        ...,
        examples=[
            [
                {
                    "aggregation": "SUM",
                    "col_id": "abc277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Au content",
                    "output_unit_id": "kg",
                }
            ]
        ],
        title="Columns",
    )
    """
    Columns within this Report Specification
    """
    cutoff_col_id: UUID | None = Field(None, examples=["13c277c2-edc6-4a7a-8380-251dd19231f2"], title="Cutoff Col Id")
    """
    ID of the column to use for cut-off evaluation
    """
    cutoff_values: list[StrictFloat] | None = Field(
        None,
        examples=[[0.5, 1.5, 5]],
        max_length=20,
        min_length=1,
        title="Cutoff Values",
    )
    """
    Values to use for cut-off evaluation
    """
    density_col_id: UUID | None = Field(None, examples=["24c277c2-edc6-4a7a-8380-251dd19231f2"], title="Density Col Id")
    """
    ID of the column to use for block density
    """
    density_unit_id: StrictStr | None = Field(None, examples=["kg/m3"], title="Density Unit Id")
    """
    ID of the unit to use for block density. The unit must be of type `MASS_PER_VOLUME`.
    """
    density_value: StrictFloat | None = Field(None, examples=[2.5], gt=1, title="Density Value")
    """
    Value to use for block density
    """
    description: StrictStr | None = Field(
        None,
        examples=["Gold resource report for test purposes"],
        max_length=1000,
        title="Description",
    )
    """
    User-supplied description of the report
    """
    mass_unit_id: StrictStr = Field(..., examples=["t"], title="Mass Unit Id")
    """
    ID of the unit to use for total mass. The unit must be of type `MASS`
    """
    name: StrictStr = Field(..., title="Name")
    """
    The human-readable label used to identify the report
    """
    negative_values_policy: ReportNegativeValuesPolicy | None = Field("IGNORE_BLOCK", examples=["USE"])
    """
    Policy for handling negative values in the report's cut-off or value columns
    """
    null_values_policy: ReportNullValuesPolicy | None = Field("IGNORE_BLOCK", examples=["ZERO"])
    """
    Policy for handling null values in the report's cut-off or value columns
    """


class DeltaRequestData(CustomBaseModel):
    bbox: BBox | BBoxXYZ = Field(..., title="Bbox")
    """
    Bounding box within which changes are searched for.
    """
    columns: list[StrictStr] = Field(..., examples=[["*"]], title="Columns")
    """
    List of column IDs to consider when checking for column updates or deletions. Also accepts `['*']`, which will consider all columns.
    """
    end_version_uuid: UUID | None = Field(None, title="End Version Uuid")
    """
    Last version to search for changes in. Any versions newer than this are not searched for changes. If not supplied, changes are searched for within all version beyond the start version, up to and including the latest version of the block model.
    """
    verbose: StrictBool = Field(False, title="Verbose")
    """
    If there are changes, whether to return details about such changes or not.
    """


class Mapping(CustomBaseModel):
    columns: list[Column] = Field(..., title="Columns")


class QueryCriteria(CustomBaseModel):
    bbox: BBox | BBoxXYZ | None = Field(None, title="Bbox")
    """

    Bounding box of the search area as integer indexes or model aligned coordinates. If not provided, this is set to the entire block model.
        If using integer indexes, they must be `>=` zero and `<=` the number of blocks - 1.
        For example, if a block model has:
        ```
        n_blocks: {
            nx: 10,
            ny: 10,
            nz: 10
        }
        ```
        then i_max, j_max, and k_max must all be `>=` 0 and `<=` 9.
        If using model aligned coordinates, that is, rotated by the same scheme as the block model, centroids that are
        encapsulated by the bounding box will be selected. A side effect of this behaviour is a bounding box with x, y, z
        min-max can exceed the model along any axis.


    """
    columns: list[StrictStr] = Field(..., examples=[["*"]], title="Columns")
    """

    List of columns, in addition to the "geometry" columns, that will be included in the output file.
    The `columns` field supports selecting columns by either their title or ID, and can also include a wildcard (`"*"`) placeholder, which will expand to all user columns, ordered alphabetically by title in a case-insensitive manner.
    Please note that the wildcard does not cover the system column `version_id`. To include `version_id` in the output file, it must also be explicitly specified in the `columns` field alongside the wildcard.
    The order of columns in the output file will match the order in the `columns` field. Columns that are part of the initial "geometry" columns will be ignored if specified.

    """
    geometry_columns: GeometryColumns = "indices"
    """

    Determines whether the blocks in the output file will be primarily identified by their coordinates or their block
    indices.

    If `geometry_columns` is set to "indices" (the default), then the first columns in the output file will be `i`, `j`, `k` followed by the sub-block index columns (if applicable).
    If `geometry_columns` is set to "coordinates", then the output file will contain the columns `x`, `y`, `z` for regular block models and `x`, `y`, `z`, `dx`, `dy`, `dz` for sub-blocked block models. These columns are referred to as the "geometry" columns, and will appear before any others columns specified in the `columns` field.

    """
    output_options: OutputOptionsParquet | OutputOptionsCSV | None = Field(
        None, discriminator="file_format", title="Output Options"
    )
    """
    Format of the output file
    """
    version_uuid: UUID | None = Field(None, title="Version Uuid")
    """
    Version UUID of the version of the block model to query, or the latest version if not provided
    """


class QueryResult(CustomBaseModel):
    bbox: BBox | BBoxXYZ = Field(..., title="Bbox")
    """

    Bounding box of the search area as integer indexes or model aligned coordinates. If not provided, this is set to the entire block model.
        If using integer indexes, they must be `>=` zero and `<=` the number of blocks - 1.
        For example, if a block model has:
        ```
        n_blocks: {
            nx: 10,
            ny: 10,
            nz: 10
        }
        ```
        then i_max, j_max, and k_max must all be `>=` 0 and `<=` 9.
        If using model aligned coordinates, that is, rotated by the same scheme as the block model, centroids that are
        encapsulated by the bounding box will be selected. A side effect of this behaviour is a bounding box with x, y, z
        min-max can exceed the model along any axis.


    """
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model
    """
    columns: list[StrictStr] = Field(..., title="Columns")
    """

    List of columns, in addition to the "geometry" columns, that will be included in the output file.
    The `columns` field supports selecting columns by either their title or ID, and can also include a wildcard (`"*"`) placeholder, which will expand to all user columns, ordered alphabetically by title in a case-insensitive manner.
    Please note that the wildcard does not cover the system column `version_id`. To include `version_id` in the output file, it must also be explicitly specified in the `columns` field alongside the wildcard.
    The order of columns in the output file will match the order in the `columns` field. Columns that are part of the initial "geometry" columns will be ignored if specified.

    """
    job_url: AnyUrl = Field(..., title="Job Url")
    """
    Pollable job URL for this query
    """
    mapping: Mapping
    """

    List of columns, in addition to the "geometry" columns, that will be included in the output file.
    The `columns` field supports selecting columns by either their title or ID, and can also include a wildcard (`"*"`) placeholder, which will expand to all user columns, ordered alphabetically by title in a case-insensitive manner.
    Please note that the wildcard does not cover the system column `version_id`. To include `version_id` in the output file, it must also be explicitly specified in the `columns` field alongside the wildcard.
    The order of columns in the output file will match the order in the `columns` field. Columns that are part of the initial "geometry" columns will be ignored if specified.

    """
    version_id: StrictInt = Field(..., examples=[5], title="Version Id")
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """
    version_uuid: UUID = Field(..., examples=["3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7"], title="Version Uuid")
    """
    A universally unique identifier for the version
    """


class ReportComparisonResultInfo(CustomBaseModel):
    referenced_columns: list[ReportResultReferencedColumn] = Field(..., title="Referenced Columns")
    report_result_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Report Result Created At")
    """
    Creation date of the Report Result
    """
    report_result_uuid: UUID = Field(
        ...,
        examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Result Uuid",
    )
    """
    ID of the Report Result
    """
    version_comment: StrictStr | None = Field(None, examples=["Updated density"], title="Version Comment")
    """
    Comment of the version of the block model that the report was run on
    """
    version_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Version Created At")
    """
    Creation date of the version of the block model that the report was run on
    """
    version_created_by: UserInfo
    """
    User who created the version of the block model that the report was run on
    """
    version_id: StrictInt = Field(..., examples=[2], title="Version Id")
    """
    Version ID of the version of the block model that the report was run on
    """
    version_uuid: UUID | None = Field(None, examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"], title="Version Uuid")
    """
    Version UUID of the version of block model that the report was run on
    """


class ReportComparisonRow(CustomBaseModel):
    categories: list[StrictStr | StrictInt | StrictFloat | StrictBool | None] = Field(
        ..., examples=[["Granite"]], title="Categories"
    )
    """
    List of category values. `null` indicates a total for that category column.
    """
    values: list[ReportComparisonValue] = Field(..., examples=[[2.7, 0.3]], title="Values")
    """
    List of values for the value columns
    """


class ReportResultSet(CustomBaseModel):
    cutoff_value: StrictFloat | None = Field(None, examples=[0.5], title="Cutoff Value")
    """
    Cut-off value for this set
    """
    rows: list[ReportRow] = Field(..., title="Rows")
    """
    List of rows in this result set
    """


class ReportResultSummary(CustomBaseModel):
    report_result_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Report Result Created At")
    """
    Creation date of the Report Result
    """
    report_result_uuid: UUID = Field(
        ...,
        examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Result Uuid",
    )
    """
    ID of the Report Result
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the report specification that was run
    """
    version_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Version Created At")
    """
    Creation date of the version of the block model that the report was run on
    """
    version_created_by: UserInfo
    """
    User who created the version of the block model that the report was run on
    """
    version_id: StrictInt = Field(..., examples=[2], title="Version Id")
    """
    Version ID of the version of block model that the report was run on
    """
    version_uuid: UUID | None = Field(None, examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"], title="Version Uuid")
    """
    Version UUID of the version of block model that the report was run on
    """


class ReportSpecificationWithJobUrl(CustomBaseModel):
    autorun: StrictBool = Field(True, title="Autorun")
    """
    Whether to automatically run this Report Specification when a new version is published
    """
    bbox: BBox | BBoxXYZ | None = Field(
        None,
        examples=[
            {
                "i_minmax": {"max": 1, "min": 0},
                "j_minmax": {"max": 1, "min": 0},
                "k_minmax": {"max": 1, "min": 0},
            }
        ],
        title="Bbox",
    )
    """
    Bounding box for the report
    """
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model this Report Specification is associated with
    """
    categories: list[ReportCategory] | None = Field(
        None,
        examples=[
            [
                {
                    "col_id": "11c277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Grade",
                    "values": ["low", "medium", "high"],
                }
            ]
        ],
        max_length=1,
        min_length=0,
        title="Categories",
    )
    """
    Category columns within this Report Specification. If null or empty, the report will have a single total row for each cut-off.
    """
    columns: list[ReportColumn] = Field(
        ...,
        examples=[
            [
                {
                    "aggregation": "SUM",
                    "col_id": "abc277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Au content",
                    "output_unit_id": "kg",
                }
            ]
        ],
        title="Columns",
    )
    """
    Columns within this Report Specification
    """
    cutoff_col_id: UUID | None = Field(None, examples=["13c277c2-edc6-4a7a-8380-251dd19231f2"], title="Cutoff Col Id")
    """
    ID of the column to use for cut-off evaluation
    """
    cutoff_values: list[StrictFloat] | None = Field(
        None,
        examples=[[0.5, 1.5, 5]],
        max_length=20,
        min_length=1,
        title="Cutoff Values",
    )
    """
    Values to use for cut-off evaluation
    """
    density_col_id: UUID | None = Field(None, examples=["24c277c2-edc6-4a7a-8380-251dd19231f2"], title="Density Col Id")
    """
    ID of the column to use for block density
    """
    density_unit_id: StrictStr | None = Field(None, examples=["kg/m3"], title="Density Unit Id")
    """
    ID of the unit to use for block density. The unit must be of type `MASS_PER_VOLUME`.
    """
    density_value: StrictFloat | None = Field(None, examples=[2.5], gt=1, title="Density Value")
    """
    Value to use for block density
    """
    description: StrictStr | None = Field(
        None,
        examples=["Gold resource report for test purposes"],
        max_length=1000,
        title="Description",
    )
    """
    User-supplied description of the report
    """
    job_url: AnyUrl | None = Field(None, title="Job Url")
    """
    URL for the Reporting Job, if `run_now` was requested
    """
    mass_unit_id: StrictStr = Field(..., examples=["t"], title="Mass Unit Id")
    """
    ID of the unit to use for total mass. The unit must be of type `MASS`
    """
    name: StrictStr = Field(..., title="Name")
    """
    The human-readable label used to identify the report
    """
    negative_values_policy: ReportNegativeValuesPolicy | None = Field("IGNORE_BLOCK", examples=["USE"])
    """
    Policy for handling negative values in the report's cut-off or value columns
    """
    null_values_policy: ReportNullValuesPolicy | None = Field("IGNORE_BLOCK", examples=["ZERO"])
    """
    Policy for handling null values in the report's cut-off or value columns
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the Report Specification
    """
    revision: StrictInt = Field(0, title="Revision")
    """
    Revision number of this Report Specification
    """


class ReportSpecificationWithLastRunInfo(CustomBaseModel):
    autorun: StrictBool = Field(True, title="Autorun")
    """
    Whether to automatically run this Report Specification when a new version is published
    """
    bbox: BBox | BBoxXYZ | None = Field(
        None,
        examples=[
            {
                "i_minmax": {"max": 1, "min": 0},
                "j_minmax": {"max": 1, "min": 0},
                "k_minmax": {"max": 1, "min": 0},
            }
        ],
        title="Bbox",
    )
    """
    Bounding box for the report
    """
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model this Report Specification is associated with
    """
    categories: list[ReportCategory] | None = Field(
        None,
        examples=[
            [
                {
                    "col_id": "11c277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Grade",
                    "values": ["low", "medium", "high"],
                }
            ]
        ],
        max_length=1,
        min_length=0,
        title="Categories",
    )
    """
    Category columns within this Report Specification. If null or empty, the report will have a single total row for each cut-off.
    """
    columns: list[ReportColumn] = Field(
        ...,
        examples=[
            [
                {
                    "aggregation": "SUM",
                    "col_id": "abc277c2-edc6-4a7a-8380-251dd19231f2",
                    "label": "Au content",
                    "output_unit_id": "kg",
                }
            ]
        ],
        title="Columns",
    )
    """
    Columns within this Report Specification
    """
    cutoff_col_id: UUID | None = Field(None, examples=["13c277c2-edc6-4a7a-8380-251dd19231f2"], title="Cutoff Col Id")
    """
    ID of the column to use for cut-off evaluation
    """
    cutoff_values: list[StrictFloat] | None = Field(
        None,
        examples=[[0.5, 1.5, 5]],
        max_length=20,
        min_length=1,
        title="Cutoff Values",
    )
    """
    Values to use for cut-off evaluation
    """
    density_col_id: UUID | None = Field(None, examples=["24c277c2-edc6-4a7a-8380-251dd19231f2"], title="Density Col Id")
    """
    ID of the column to use for block density
    """
    density_unit_id: StrictStr | None = Field(None, examples=["kg/m3"], title="Density Unit Id")
    """
    ID of the unit to use for block density. The unit must be of type `MASS_PER_VOLUME`.
    """
    density_value: StrictFloat | None = Field(None, examples=[2.5], gt=1, title="Density Value")
    """
    Value to use for block density
    """
    description: StrictStr | None = Field(
        None,
        examples=["Gold resource report for test purposes"],
        max_length=1000,
        title="Description",
    )
    """
    User-supplied description of the report
    """
    last_result_created_at: datetime | None = Field(
        None, examples=["2021-01-01T00:00:00Z"], title="Last Result Created At"
    )
    """
    Creation date of the Report Result
    """
    last_result_version_id: StrictInt | None = Field(None, examples=[2], title="Last Result Version Id")
    """
    Version ID of the version of the block model that the report was run on
    """
    mass_unit_id: StrictStr = Field(..., examples=["t"], title="Mass Unit Id")
    """
    ID of the unit to use for total mass. The unit must be of type `MASS`
    """
    name: StrictStr = Field(..., title="Name")
    """
    The human-readable label used to identify the report
    """
    negative_values_policy: ReportNegativeValuesPolicy | None = Field("IGNORE_BLOCK", examples=["USE"])
    """
    Policy for handling negative values in the report's cut-off or value columns
    """
    null_values_policy: ReportNullValuesPolicy | None = Field("IGNORE_BLOCK", examples=["ZERO"])
    """
    Policy for handling null values in the report's cut-off or value columns
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the Report Specification
    """
    revision: StrictInt = Field(0, title="Revision")
    """
    Revision number of this Report Specification
    """


class ReportWarning(CustomBaseModel):
    message: StrictStr = Field(
        ...,
        examples=["Density value was null, or less than or equal to 0. The default density value of 2.7 was used."],
        title="Message",
    )
    """
    Warning message
    """
    occurrences: StrictInt = Field(..., examples=[1], title="Occurrences")
    """
    Number of occurrences of the warning
    """
    warning_type: ReportWarningType = Field(..., examples=["INVALID_DENSITY_VALUE"])
    """
    Type of warning
    """


class Rotation(CustomBaseModel):
    angle: StrictFloat = Field(..., title="Angle")
    """
    Angle of rotation in degrees
    """
    axis: RotationAxis
    """
    Axis of rotation
    """


class Unit(CustomBaseModel):
    conversion_factor: StrictFloat = Field(..., examples=[1], title="Conversion Factor")
    """
    Conversion factor to convert to the reference unit for this unit type
    """
    description: StrictStr = Field(..., examples=["Metres"], title="Description")
    """
    Description of the unit
    """
    symbol: StrictStr = Field(..., examples=["m"], title="Symbol")
    """
    Display symbol for the unit
    """
    unit_id: StrictStr = Field(..., examples=["m"], title="Unit Id")
    """
    ID of the unit
    """
    unit_type: UnitType = Field(..., examples=["LENGTH"])
    """
    Type of the unit
    """


class UpdateMetadata(CustomBaseModel):
    col_id: StrictStr = Field(..., examples=["618d6339-2fa7-4dfd-9c7f-c0b12016639e"], title="Col Id")
    """
    The ID of the column, a UUID for non-system columns
    """
    values: UpdateMetadataValues


class UpdateMetadataLite(CustomBaseModel):
    title: StrictStr = Field(..., title="Title")
    """
    The original human-readable label used to identify the column
    """
    values: UpdateMetadataValues


class Version(CustomBaseModel):
    """
    API output
    """

    base_version_id: StrictInt | None = Field(None, examples=[4], title="Base Version Id")
    """
    Version the update was applied to. This will be the same as `parent_version_id`, except for
    updates made by Leapfrog, where it is the current local version when the block model is published. This is null if this
    is the first version.
    """
    bbox: BBox | None = None
    """

    Bounding box of data updated between this version and last version. Will be None for the initial version, and updates
    that only delete and rename columns.

    """
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model
    """
    comment: StrictStr = Field("", title="Comment")
    """
    User-supplied comment
    """
    created_at: datetime = Field(..., title="Created At")
    """
    When the version was created
    """
    created_by: UserInfo
    """
    User who performed the action that created the version
    """
    geoscience_version_id: StrictStr | None = Field(None, examples=["1234567890"], title="Geoscience Version Id")
    """
    ID of the Geoscience Object Service object version associated with this block model version
    """
    mapping: Mapping
    """
    Columns within this version
    """
    parent_version_id: StrictInt | None = Field(None, examples=[4], title="Parent Version Id")
    """
    Previous version. 0 if this is the first version.
    """
    version_id: StrictInt = Field(..., examples=[5], title="Version Id")
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """
    version_uuid: UUID = Field(..., examples=["3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7"], title="Version Uuid")
    """
    A universally unique identifier for the version
    """


class VersionColumnChanges(CustomBaseModel):
    deleted: list[StrictStr] = Field([], title="Deleted")
    metadata_updated: list[UpdateMetadataLite] | None = Field([], title="Metadata Updated")
    new: list[StrictStr] = Field([], title="New")
    renamed: list[ColumnRename] = Field([], title="Renamed")
    updated: list[StrictStr] = Field([], title="Updated")


class BlockModel(CustomBaseModel):
    bbox: BBoxXYZ
    """

    Axis-aligned bounding box of the block model.

    This is the smallest box that fully contains all blocks within the block model, regardless of whether they contain data.
    It is defined by the minimum and maximum coordinates along each axis.

    """
    block_rotation: list[Rotation] = Field(..., title="Block Rotation")
    """

    The rotation of the block model. Defined by a list of clockwise rotations around given axes.

    The combined equation for the rotation is given by:

        R_1 ... R_n A = B

    where `A` is the location in the IJK space, `B` is the location in coordinate space, and `R_i` is the rotation matrix for the `i`th rotation.

    For example, if the first rotation is a rotation of 15 degrees clockwise around the X axis, `R_1` would be `[[1, 0, 0], [0, cos(15), sin(15)], [0, -sin(15), cos(15)]]`.

    Two rotations around the same axis cannot be consecutive items within the list of rotations.

    """
    bm_uuid: UUID = Field(..., title="Block model ID")
    """
    ID of the block model
    """
    coordinate_reference_system: StrictStr | None = Field(
        None, examples=["EPSG:3395"], title="Coordinate Reference System"
    )
    """

    Coordinate reference system used in the block model.

    This may be an EPSG code, signified by the prefix 'EPSG:', or a coordinate system definition in WKT format.

    """
    created_at: datetime = Field(..., title="Created At")
    """
    Block model creation time
    """
    created_by: UserInfo
    """
    User who created the block model
    """
    description: StrictStr | None = Field(None, title="Description")
    """
    User-supplied description of the block model
    """
    geoscience_object_id: UUID | None = Field(
        None,
        examples=["f1230d69-c779-4736-9eb6-f82d49aafa9a"],
        title="Geoscience Object Id",
    )
    """
    UUID of the Geoscience Object Service object associated with the block model
    """
    last_updated_at: datetime = Field(..., title="Last Updated At")
    """
    Date and time of the last block model update, including metadata updates
    """
    last_updated_by: UserInfo
    """
    User who last updated the block model, including metadata updates
    """
    model_origin: Location
    """
    Origin of the block model
    """
    name: StrictStr = Field(..., title="Name")
    """
    The human-readable label used to identify the block model
    """
    normalized_rotation: list = Field(..., max_length=3, min_length=3, title="Normalized Rotation")
    """
    Normalised rotation of the block model, represented as intrinsic rotations around the Z axis, then X axis, then Z axis.
    """
    org_uuid: UUID = Field(..., examples=["e6230d69-c779-4736-9eb6-f82d49aafa06"], title="Organisation ID")
    """
    Organisation the block model belongs to.
    """
    size_options: SizeOptionsRegular | SizeOptionsOctree | SizeOptionsFullySubBlocked | SizeOptionsFlexible = Field(
        ..., discriminator="model_type", title="Size Options"
    )
    """
    Size of the block model
    """
    size_unit_id: StrictStr | None = Field(None, title="Size Unit Id")
    """
    Unit ID denoting the length unit used for the block model's blocks.
    """
    workspace_id: UUID = Field(..., examples=["860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef"], title="Workspace ID")
    """
    Workspace the block model belongs to.
    """


class BlockModelAndJobURL(CustomBaseModel):
    bbox: BBoxXYZ
    """

    Axis-aligned bounding box of the block model.

    This is the smallest box that fully contains all blocks within the block model, regardless of whether they contain data.
    It is defined by the minimum and maximum coordinates along each axis.

    """
    block_rotation: list[Rotation] = Field(..., title="Block Rotation")
    """

    The rotation of the block model. Defined by a list of clockwise rotations around given axes.

    The combined equation for the rotation is given by:

        R_1 ... R_n A = B

    where `A` is the location in the IJK space, `B` is the location in coordinate space, and `R_i` is the rotation matrix for the `i`th rotation.

    For example, if the first rotation is a rotation of 15 degrees clockwise around the X axis, `R_1` would be `[[1, 0, 0], [0, cos(15), sin(15)], [0, -sin(15), cos(15)]]`.

    Two rotations around the same axis cannot be consecutive items within the list of rotations.

    """
    bm_uuid: UUID = Field(..., title="Block model ID")
    """
    ID of the block model
    """
    coordinate_reference_system: StrictStr | None = Field(
        None, examples=["EPSG:3395"], title="Coordinate Reference System"
    )
    """

    Coordinate reference system used in the block model.

    This may be an EPSG code, signified by the prefix 'EPSG:', or a coordinate system definition in WKT format.

    """
    created_at: datetime = Field(..., title="Created At")
    """
    Block model creation time
    """
    created_by: UserInfo
    """
    User who created the block model
    """
    description: StrictStr | None = Field(None, title="Description")
    """
    User-supplied description of the block model
    """
    geoscience_object_id: UUID | None = Field(
        None,
        examples=["f1230d69-c779-4736-9eb6-f82d49aafa9a"],
        title="Geoscience Object Id",
    )
    """
    UUID of the Geoscience Object Service object associated with the block model
    """
    job_url: AnyUrl = Field(..., title="Job Url")
    last_updated_at: datetime = Field(..., title="Last Updated At")
    """
    Date and time of the last block model update, including metadata updates
    """
    last_updated_by: UserInfo
    """
    User who last updated the block model, including metadata updates
    """
    model_origin: Location
    """
    Origin of the block model
    """
    name: StrictStr = Field(..., title="Name")
    """
    The human-readable label used to identify the block model
    """
    normalized_rotation: list = Field(..., max_length=3, min_length=3, title="Normalized Rotation")
    """
    Normalised rotation of the block model, represented as intrinsic rotations around the Z axis, then X axis, then Z axis.
    """
    org_uuid: UUID = Field(..., examples=["e6230d69-c779-4736-9eb6-f82d49aafa06"], title="Organisation ID")
    """
    Organisation the block model belongs to.
    """
    size_options: SizeOptionsRegular | SizeOptionsOctree | SizeOptionsFullySubBlocked | SizeOptionsFlexible = Field(
        ..., discriminator="model_type", title="Size Options"
    )
    """
    Size of the block model
    """
    size_unit_id: StrictStr | None = Field(None, title="Size Unit Id")
    """
    Unit ID denoting the length unit used for the block model's blocks.
    """
    workspace_id: UUID = Field(..., examples=["860be2f5-fe06-4c1b-ac8b-7d34d2b6d2ef"], title="Workspace ID")
    """
    Workspace the block model belongs to.
    """


class CreateData(CustomBaseModel):
    block_rotation: list[Rotation] = Field(..., max_length=3, title="Block Rotation")
    """

    The rotation of the block model. Defined by a list of clockwise rotations around given axes.

    The combined equation for the rotation is given by:

        R_1 ... R_n A = B

    where `A` is the location in the IJK space, `B` is the location in coordinate space, and `R_i` is the rotation matrix for the `i`th rotation.

    For example, if the first rotation is a rotation of 15 degrees clockwise around the X axis, `R_1` would be `[[1, 0, 0], [0, cos(15), sin(15)], [0, -sin(15), cos(15)]]`.

    Two rotations around the same axis cannot be consecutive items within the list of rotations.

    """
    comment: StrictStr | None = Field(None, max_length=250, title="Comment")
    coordinate_reference_system: StrictStr | None = Field(
        None,
        examples=["EPSG:3395"],
        max_length=10000,
        title="Coordinate Reference System",
    )
    """

    Coordinate reference system used in the block model.

    This may be an EPSG code, signified by the prefix 'EPSG:', or a coordinate system definition in WKT format.

    """
    description: StrictStr | None = Field(None, max_length=500, title="Description")
    """
    Description of the block model
    """
    model_origin: Location
    name: StrictStr = Field(..., title="Name")
    r"""
    Name of the block model. This may not contain `/` nor `\`.
    """
    object_path: StrictStr | None = Field(None, examples=["/path/to/folder"], title="Object Path")
    r"""

    Path of the folder in Geoscience Object Service to create the reference object in.

    This may not contain relative elements like `.` or `..`. It may not contain double slashes `//`, nor backslashes `\`,
    nor elements ending in `.`.

    """
    size_options: SizeOptionsRegular | SizeOptionsOctree | SizeOptionsFullySubBlocked | SizeOptionsFlexible = Field(
        ..., discriminator="model_type", title="Size Options"
    )
    size_unit_id: StrictStr | None = Field(None, examples=["m"], title="Size Unit Id")
    """
    Unit ID denoting the length unit used for the block model's blocks.
    """


class JobResponse(CustomBaseModel):
    job_status: JobStatus = Field(..., examples=["COMPLETE"])
    """

    Status of this job. The statuses have the following meanings:
     - `PENDING_UPLOAD`: This is a model update job, and is awaiting notification that file upload has
       completed before proceeding any further.
     - `QUEUED`: The job is ready for processing, and is queued waiting for a worker process to become available.
     - `PROCESSING`: The job is actively being processed by a worker.
     - `COMPLETE`: The job completed successfully.
     - `FAILED`: The job failed. No further processing of this job will be attempted.

    """
    payload: (
        Version
        | QueryDownload
        | JobErrorPayload
        | ReportRunResult
        | ReportComparisonJobResult
        | JobPendingPayload
        | None
    ) = Field(None, title="Payload")
    """

    The payload describes the outcome of the job, and is set for completed or failed jobs. This property will not appear if the
    job is not yet `COMPLETE` or `FAILED`.
     - For all failed jobs, this will be a `JobErrorPayload`.
     - For completed create or update jobs, this will be a `Version`.
     - For completed query jobs, this will be a `QueryDownload`.

    """


class PaginatedResponseWithUnitsReportSpecificationWithLastRunInfo(CustomBaseModel):
    count: StrictInt = Field(..., title="Count")
    """
    Number of results returned in `results`
    """
    limit: StrictInt = Field(..., title="Limit")
    """
    Maximum number of items requested
    """
    offset: StrictInt = Field(..., title="Offset")
    """
    Index of the first item in `results` with respect to the full list without pagination
    """
    referenced_units: list[Unit] = Field(..., title="Referenced Units")
    """
    List of all units referenced in the results
    """
    results: list[ReportSpecificationWithLastRunInfo] = Field(..., title="Results")
    """
    List of results
    """
    total: StrictInt = Field(..., title="Total")
    """
    Total number of items within the full list without pagination
    """


class PaginatedResponseWithUnitsVersion(CustomBaseModel):
    count: StrictInt = Field(..., title="Count")
    """
    Number of results returned in `results`
    """
    limit: StrictInt = Field(..., title="Limit")
    """
    Maximum number of items requested
    """
    offset: StrictInt = Field(..., title="Offset")
    """
    Index of the first item in `results` with respect to the full list without pagination
    """
    referenced_units: list[Unit] = Field(..., title="Referenced Units")
    """
    List of all units referenced in the results
    """
    results: list[Version] = Field(..., title="Results")
    """
    List of results
    """
    total: StrictInt = Field(..., title="Total")
    """
    Total number of items within the full list without pagination
    """


class PaginatedResponseBlockModel(CustomBaseModel):
    count: StrictInt = Field(..., title="Count")
    """
    Number of results returned in `results`
    """
    limit: StrictInt = Field(..., title="Limit")
    """
    Maximum number of items requested
    """
    offset: StrictInt = Field(..., title="Offset")
    """
    Index of the first item in `results` with respect to the full list without pagination
    """
    results: list[BlockModel] = Field(..., title="Results")
    """
    List of results
    """
    total: StrictInt = Field(..., title="Total")
    """
    Total number of items within the full list without pagination
    """


class PaginatedResponseReportResultSummary(CustomBaseModel):
    count: StrictInt = Field(..., title="Count")
    """
    Number of results returned in `results`
    """
    limit: StrictInt = Field(..., title="Limit")
    """
    Maximum number of items requested
    """
    offset: StrictInt = Field(..., title="Offset")
    """
    Index of the first item in `results` with respect to the full list without pagination
    """
    results: list[ReportResultSummary] = Field(..., title="Results")
    """
    List of results
    """
    total: StrictInt = Field(..., title="Total")
    """
    Total number of items within the full list without pagination
    """


class ReportComparisonResultSet(CustomBaseModel):
    cutoff_value: StrictFloat | None = Field(None, examples=[1], title="Cutoff Value")
    """
    Cut-off value used in the report
    """
    rows: list[ReportComparisonRow] = Field(
        ..., examples=[{"categories": ["Granite"], "values": [2.7, 0.3]}], title="Rows"
    )
    """
    List of rows in the Report Result
    """


class ReportComparisonWarnings(CustomBaseModel):
    comparison: list[ReportWarning] = Field(..., title="Comparison")
    from_result: list[ReportWarning] = Field(..., title="From Result")
    to_result: list[ReportWarning] = Field(..., title="To Result")


class ReportResult(CustomBaseModel):
    bm_uuid: UUID = Field(..., examples=["bbb277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model that the report was run on
    """
    categories: list[ReportResultCategory] = Field(
        ...,
        examples=[{"col_id": "ddd277c2-edc6-4a7a-8380-251dd19231f2", "label": "Rock type"}],
        title="Categories",
    )
    """
    List of categories used in the report
    """
    cutoff_col_id: UUID | None = Field(None, examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"], title="Cutoff Col Id")
    """
    ID of the column that was used for cut-off evaluation
    """
    referenced_columns: list[ReportResultReferencedColumn] = Field(..., title="Referenced Columns")
    report_result_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Report Result Created At")
    """
    Creation date of the Report Result
    """
    report_result_uuid: UUID = Field(
        ...,
        examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Result Uuid",
    )
    """
    ID of the Report Result
    """
    report_specification_description: StrictStr | None = Field(
        None,
        examples=["A report to show the density of the block model"],
        title="Report Specification Description",
    )
    """
    Description of the report specification that was run
    """
    report_specification_name: StrictStr = Field(..., examples=["Density Report"], title="Report Specification Name")
    """
    Name of the report specification that was run
    """
    report_specification_revision: StrictInt = Field(..., examples=[1], title="Report Specification Revision")
    """
    Revision of the report specification that was run
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the report specification that was run
    """
    result_sets: list[ReportResultSet] = Field(..., title="Result Sets")
    value_columns: list[ReportResultColumn] = Field(
        ...,
        examples=[
            {
                "col_id": "eee277c2-edc6-4a7a-8380-251dd19231f2",
                "label": "Average gold content",
                "unit_id": "ppm[mass]",
            }
        ],
        title="Value Columns",
    )
    """
    List of value columns used in the report
    """
    version_comment: StrictStr | None = Field(None, examples=["Updated density"], title="Version Comment")
    """
    Comment of the version of the block model that the report was run on
    """
    version_created_at: datetime = Field(..., examples=["2021-01-01T00:00:00Z"], title="Version Created At")
    """
    Creation date of the version of the block model that the report was run on
    """
    version_created_by: UserInfo
    """
    User who created the version of the block model that the report was run on
    """
    version_id: StrictInt = Field(..., examples=[2], title="Version Id")
    """
    Version ID of the version of block model that the report was run on
    """
    version_uuid: UUID | None = Field(None, examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"], title="Version Uuid")
    """
    Version UUID of the version of block model that the report was run on
    """
    warnings: list[ReportWarning] = Field(..., title="Warnings")


class UpdateColumns(CustomBaseModel):
    delete: list[StrictStr] = Field(..., title="Delete")
    """
    The list of column IDs for deletion
    """
    new: list[Column] = Field(..., title="New")
    rename: list[Rename] = Field(..., title="Rename")
    update: list[StrictStr] = Field(..., title="Update")
    """
    The list of column IDs for update
    """
    update_metadata: list[UpdateMetadata] | None = Field(None, title="Update Metadata")


class UpdateColumnsLiteInput(CustomBaseModel):
    delete: list[StrictStr] = Field(..., title="Delete")
    """
    The list of column titles for deletion
    """
    new: list[ColumnLite] = Field(..., title="New")
    rename: list[RenameLite] = Field(..., title="Rename")
    update: list[StrictStr] = Field(..., title="Update")
    """
    The list of column titles for update
    """
    update_metadata: list[UpdateMetadataLite] | None = Field(None, title="Update Metadata")


class UpdateColumnsLiteOutput(CustomBaseModel):
    delete: list[StrictStr] = Field(..., title="Delete")
    """
    The list of column titles for deletion
    """
    new: list[ColumnLite] = Field(..., title="New")
    rename: list[RenameLite] = Field(..., title="Rename")
    update: list[StrictStr] = Field(..., title="Update")
    """
    The list of column titles for update
    """
    update_metadata: list[UpdateMetadataLite] | None = Field(None, title="Update Metadata")


class UpdateDataLiteInput(CustomBaseModel):
    columns: UpdateColumnsLiteInput
    comment: StrictStr | None = Field(None, max_length=250, title="Comment")
    geometry_change: StrictBool | None = Field(None, title="Geometry Change")
    """

    Whether the update will create or destroy sub-blocks.
    - If set to `false`, no sub-blocks can be created nor destroyed, only sub-blocks that exist in the latest version can be updated.
    - If set to `true`:
        - New sub-blocks can be created and/or existing sub-blocks can be destroyed, but all columns must be provided.
        - Only `new`, `update`, and `delete` column operations are allowed as standalone operations.
        - `update_metadata` is allowed, but only when the target columns are being updated as part of the `update` operation, and only when there are no columns being renamed.
        - If `update_type` is set to `merge`, then all sub-blocks within a parent block must be provided.

    """
    input_options: InputOptionsParquet | InputOptionsCSV | None = Field(
        None, discriminator="file_format", title="Input Options"
    )
    update_type: UpdateType = "merge"
    """

    Behaviour of the update, for blocks that are omitted from the update file.
    - If set to `replace`, the values for blocks that are omitted will be set to null, for the selected columns. If `geometry_change` is true, the geometry of any parent block that has no blocks in the update file will be reset to an un-subdivided state.
    - If set to `merge`, blocks that are omitted will be left unchanged, the values within those blocks will be the same as within the previous version.

    """


class UpdateDataLiteOutput(CustomBaseModel):
    columns: UpdateColumnsLiteOutput
    comment: StrictStr | None = Field(None, max_length=250, title="Comment")
    geometry_change: StrictBool | None = Field(None, title="Geometry Change")
    """

    Whether the update will create or destroy sub-blocks.
    - If set to `false`, no sub-blocks can be created nor destroyed, only sub-blocks that exist in the latest version can be updated.
    - If set to `true`:
        - New sub-blocks can be created and/or existing sub-blocks can be destroyed, but all columns must be provided.
        - Only `new`, `update`, and `delete` column operations are allowed as standalone operations.
        - `update_metadata` is allowed, but only when the target columns are being updated as part of the `update` operation, and only when there are no columns being renamed.
        - If `update_type` is set to `merge`, then all sub-blocks within a parent block must be provided.

    """
    input_options: InputOptionsParquet | InputOptionsCSV | None = Field(
        None, discriminator="file_format", title="Input Options"
    )
    update_type: UpdateType = UpdateType.merge
    """

    Behaviour of the update, for blocks that are omitted from the update file.
    - If set to `replace`, the values for blocks that are omitted will be set to null, for the selected columns. If `geometry_change` is true, the geometry of any parent block that has no blocks in the update file will be reset to an un-subdivided state.
    - If set to `merge`, blocks that are omitted will be left unchanged, the values within those blocks will be the same as within the previous version.

    """


class UpdateDataWithVersion(CustomBaseModel):
    base_version_uuid: UUID = Field(..., title="Base Version Uuid")
    bbox: BBox | BBoxXYZ | None = Field(None, title="Bbox")
    columns: UpdateColumns
    comment: StrictStr | None = Field(None, max_length=250, title="Comment")
    geometry_change: StrictBool | None = Field(None, title="Geometry Change")
    """

    Whether the update will create or destroy sub-blocks.
    - If set to `false`, no sub-blocks can be created nor destroyed, only sub-blocks that exist in the latest version can be updated.
    - If set to `true`:
        - New sub-blocks can be created and/or existing sub-blocks can be destroyed, but all columns must be provided.
        - Only `new`, `update`, and `delete` column operations are allowed as standalone operations.
        - `update_metadata` is allowed, but only when the target columns are being updated as part of the `update` operation, and only when there are no columns being renamed.
        - If `update_type` is set to `merge`, then all sub-blocks within a parent block must be provided.

    """
    update_type: UpdateType = "merge"
    """

    Behaviour of the update, for blocks that are omitted from the update file.
    - If set to `replace`, the values for blocks that are omitted will be set to null, for the selected columns. If `geometry_change` is true, the geometry of any parent block that has no blocks in the update file will be reset to an un-subdivided state.
    - If set to `merge`, blocks that are omitted will be left unchanged, the values within those blocks will be the same as within the previous version.

    """


class UpdateWithUrl(CustomBaseModel):
    changes: UpdateDataWithVersion | UpdateDataLiteOutput = Field(..., title="Changes")
    job_url: AnyUrl = Field(..., title="Job Url")
    job_uuid: UUID = Field(..., title="Job Uuid")
    upload_url: AnyUrl | None = Field(None, title="Upload Url")
    version_uuid: UUID = Field(..., title="Version Uuid")


class VersionChanges(CustomBaseModel):
    blocks_uploaded: StrictInt | None = Field(None, title="Blocks Uploaded")
    columns: VersionColumnChanges
    total_blocks: StrictInt = Field(..., title="Total Blocks")
    update_type: UpdateType | None = None
    """

    Behaviour of the update, for blocks that are omitted from the update file.
    - If set to `replace`, the values for blocks that are omitted will be set to null, for the selected columns. If `geometry_change` is true, the geometry of any parent block that has no blocks in the update file will be reset to an un-subdivided state.
    - If set to `merge`, blocks that are omitted will be left unchanged, the values within those blocks will be the same as within the previous version.

    This is null if there were no columns updated and `geometry_change` was not set to false.

    """


class VersionWithChanges(CustomBaseModel):
    base_version_id: StrictInt | None = Field(None, examples=[4], title="Base Version Id")
    """
    Version the update was applied to. This will be the same as `parent_version_id`, except for
    updates made by Leapfrog, where it is the current local version when the block model is published. This is null if this
    is the first version.
    """
    bbox: BBox | None = None
    """

    Bounding box of data updated between this version and last version. Will be None for the initial version, and updates
    that only delete and rename columns.

    """
    bm_uuid: UUID = Field(..., examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model
    """
    changes: VersionChanges
    comment: StrictStr = Field("", title="Comment")
    """
    User-supplied comment
    """
    created_at: datetime = Field(..., title="Created At")
    """
    When the version was created
    """
    created_by: UserInfo
    """
    User who performed the action that created the version
    """
    geoscience_version_id: StrictStr | None = Field(None, examples=["1234567890"], title="Geoscience Version Id")
    """
    ID of the Geoscience Object Service object version associated with this block model version
    """
    mapping: Mapping
    """
    Columns within this version
    """
    parent_version_id: StrictInt | None = Field(None, examples=[4], title="Parent Version Id")
    """
    Previous version. 0 if this is the first version.
    """
    version_id: StrictInt = Field(..., examples=[5], title="Version Id")
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """
    version_uuid: UUID = Field(..., examples=["3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7"], title="Version Uuid")
    """
    A universally unique identifier for the version
    """


class DeltaVersionData(CustomBaseModel):
    bbox: BBox | None = None
    """
    Bounding box of the search area as integer indexes
    """
    mapping: Mapping
    """
    List of internal column names to be included into the result
    """
    update_columns: UpdateColumns
    """
    Columns added, updated, or deleted in this version
    """
    version_id: StrictInt = Field(..., examples=[5], title="Version Id")
    """
    Identifier for the version within a block model as a monotonically increasing integer, where 1 is
    the `version_id` for the version created upon creation of the block model.
    """
    version_uuid: UUID = Field(..., examples=["3e9ce8de-f6ba-4920-8c6e-0882e90f0ed7"], title="Version Uuid")
    """
    A universally unique identifier for the version
    """


class ReportComparison(CustomBaseModel):
    bm_uuid: UUID = Field(..., examples=["bbb277c2-edc6-4a7a-8380-251dd19231f2"], title="Bm Uuid")
    """
    ID of the block model that the report was run on
    """
    categories: list[ReportResultCategory] = Field(
        ...,
        examples=[{"col_id": "ddd277c2-edc6-4a7a-8380-251dd19231f2", "label": "Rock type"}],
        title="Categories",
    )
    """
    List of categories used in the report
    """
    cutoff_col_id: UUID | None = Field(None, examples=["ccc277c2-edc6-4a7a-8380-251dd19231f2"], title="Cutoff Col Id")
    """
    ID of the column that was used for cut-off evaluation
    """
    from_result: ReportComparisonResultInfo
    report_specification_description: StrictStr | None = Field(
        None,
        examples=["A report to show the density of the block model"],
        title="Report Specification Description",
    )
    """
    Description of the Report Specification that was run
    """
    report_specification_name: StrictStr = Field(..., examples=["Density Report"], title="Report Specification Name")
    """
    Name of the Report Specification that was run
    """
    report_specification_revision: StrictInt = Field(..., examples=[1], title="Report Specification Revision")
    """
    Revision of the Report Specification that was run
    """
    report_specification_uuid: UUID = Field(
        ...,
        examples=["e3c277c2-edc6-4a7a-8380-251dd19231f2"],
        title="Report Specification Uuid",
    )
    """
    ID of the Report Specification that was run
    """
    result_sets: list[ReportComparisonResultSet] = Field(..., title="Result Sets")
    to_result: ReportComparisonResultInfo
    value_columns: list[ReportResultColumn] = Field(
        ...,
        examples=[
            {
                "col_id": "eee277c2-edc6-4a7a-8380-251dd19231f2",
                "label": "Average gold content",
                "unit_id": "ppm[mass]",
            }
        ],
        title="Value Columns",
    )
    """
    List of value columns used in the report
    """
    warnings: ReportComparisonWarnings


class DeltaResponseData(CustomBaseModel):
    """
    Details about changes that occurred between two versions, within the specified bounding box.
    """

    delete_deltas: list[UUID] = Field(..., title="Delete Deltas")
    """
    List of versions that deleted specified columns
    """
    new_deltas: list[UUID] = Field(..., title="New Deltas")
    """
    List of versions that added new columns
    """
    update_deltas: list[UUID] = Field(..., title="Update Deltas")
    """
    List of versions that contain updates within the specified bounding box, to the specified columns
    """
    version_data: list[DeltaVersionData] = Field(..., title="Version Data")
    """
    Details about the changes per version
    """



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from pathlib import PurePosixPath
from uuid import UUID

from evo import logging
from evo.common import APIConnector, BaseAPIClient, Environment, HealthCheckType, Page, ServiceHealth, ServiceUser
from evo.common.utils import get_service_health

from .data import FileMetadata, FileVersion
from .endpoints import FileV2Api
from .endpoints.models import DownloadFileResponse, FileVersionResponse, ListFile, UserInfo
from .io import FileAPIDownload, FileAPIUpload

logger = logging.getLogger("file.client")

__all__ = ["FileAPIClient"]


def _user_from_model(model: UserInfo | None) -> ServiceUser | None:
    """Create a ServiceUser instance from a generated UserInfo model.

    :param model: The model to create the ServiceUser instance from, or None.

    :return: A ServiceUser instance, or None if the model is None.
    """
    return None if model is None else ServiceUser(id=model.id, name=model.name, email=model.email)


def _versions_from_listed_versions(models: list[FileVersionResponse]) -> list[FileVersion]:
    """Create a list of FileVersion instances from a list of generated FileVersionResponse models.

    :param models: The models to create the FileVersion instances from.

    :return: A sorted list of FileVersion instances.
    """
    versions = (
        FileVersion(
            version_id=model.version_id,
            created_at=model.created_at,
            created_by=_user_from_model(model.created_by),
        )
        for model in models
    )
    return sorted(versions, key=lambda v: v.created_at, reverse=True)


class FileAPIClient(BaseAPIClient):
    def __init__(self, environment: Environment, connector: APIConnector) -> None:
        """
        :param environment: The environment object
        :param connector: The connector object.
        """
        super().__init__(environment, connector)
        self._api = FileV2Api(connector=connector)

    async def get_service_health(self, check_type: HealthCheckType = HealthCheckType.FULL) -> ServiceHealth:
        """Get the health of the file service.

        :param check_type: The type of health check to perform.

        :return: A ServiceHealth object.

        :raises EvoAPIException: If the API returns an unexpected status code.
        :raises ClientValueError: If the response is not a valid service health check response.
        """
        return await get_service_health(self._connector, "file", check_type=check_type)

    def _metadata_from_listed_file(self, model: ListFile) -> FileMetadata:
        """Create a FileMetadata instance from a generated ListFile model.

        :param model: The model to create the FileMetadata instance from.

        :return: A FileMetadata instance.
        """
        return FileMetadata(
            environment=self._environment,
            id=model.file_id,
            name=model.name,
            created_at=model.created_at,
            created_by=_user_from_model(model.created_by),
            modified_at=model.modified_at,
            modified_by=_user_from_model(model.modified_by),
            parent=model.path,
            version_id=model.version_id,
            size=model.size,
        )

    def _metadata_from_endpoint_model(self, model: DownloadFileResponse) -> FileMetadata:
        """Create a FileMetadata instance from a generated DownloadFileResponse model.

        :param model: The model to create the FileMetadata instance from.

        :return: A FileMetadata instance.
        """
        file_path = PurePosixPath(model.path)
        return FileMetadata(
            environment=self._environment,
            id=model.file_id,
            name=model.name,
            created_at=model.created_at,
            created_by=_user_from_model(model.created_by),
            modified_at=model.modified_at,
            modified_by=_user_from_model(model.modified_by),
            parent=str(file_path.parent),
            version_id=model.version_id,
            size=model.size,
        )

    async def list_files(
        self,
        offset: int = 0,
        limit: int = 5000,
        name: str | None = None,
    ) -> Page[FileMetadata]:
        """List up to `limit` files in the workspace, starting at `offset`.

        The files will be the latest version of the file.
        If there are no files starting at `offset`, the page will be empty.

        :param offset: The number of files to skip before listing.
        :param limit: Max number of files to list.
        :param name: Filter files by name.

        :return: A page of all files from the query.
        """
        assert limit > 0, "Limit must be a positive integer"
        assert offset >= 0, "Offset must be a non-negative integer"
        response = await self._api.list_files(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            limit=limit,
            offset=offset,
            file_name=name,
        )
        return Page(
            offset=offset,
            limit=limit,
            total=response.total,
            items=[self._metadata_from_listed_file(file) for file in response.files],
        )

    async def list_all_files(self, limit_per_request: int = 5000, name: str | None = None) -> list[FileMetadata]:
        """List all files in the workspace.

        This method makes multiple calls to the `list_files` endpoint until all files have been listed.

        :param limit_per_request: The maximum number of files to list in one request.
        :param name: Filter files by name.

        :return: A list of all files in the workspace.
        """
        items = []
        offset = 0
        while True:
            page = await self.list_files(offset=offset, limit=limit_per_request, name=name)
            items += page.items()
            if page.is_last:
                break
            offset = page.next_offset
        return items

    async def get_file_by_path(self, path: str, version_id: str | None = None) -> FileMetadata:
        """Get a file by its path.

        :param path: The path to the file.
        :param version_id: ID of the desired file version. By default, the response will return the latest version.
        :return: A FileMetadata representation of the file on the service.
        """
        file_response = await self._api.get_file_by_path(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_path=path,
            version_id=version_id,
        )
        return self._metadata_from_endpoint_model(file_response)

    async def get_file_by_id(self, file_id: UUID, version_id: str | None = None) -> FileMetadata:
        """Get a file by its ID.

        :param file_id: UUID of a file
        :param version_id: ID of the desired file version. By default, the response will return the latest version.
        :return: A FileMetadata representation of the file on the service
        """
        file_response = await self._api.get_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(file_id),
            version_id=version_id,
        )
        return self._metadata_from_endpoint_model(file_response)

    async def list_versions_by_path(self, path: str) -> list[FileVersion]:
        """List the versions of a file by path.

        :param path: The path to the file.
        :return: A sorted list of file versions. The latest version is the first element of the list.
        """
        file_response = await self._api.get_file_by_path(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_path=path,
            include_versions=True,
        )
        return _versions_from_listed_versions(file_response.versions)

    async def list_versions_by_id(self, file_id: UUID) -> list[FileVersion]:
        """List the versions of a file by ID

        :param file_id: UUID of the file.
        :return: A sorted list of file versions. The latest version is the first element of the list.
        """
        file_response = await self._api.get_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(file_id),
            include_versions=True,
        )
        return _versions_from_listed_versions(file_response.versions)

    async def prepare_download_by_path(self, path: str, version_id: str | None = None) -> FileAPIDownload:
        """Prepares a file for download by path.

        :param path: Path to the file.
        :param version_id: Versions of the file.

        :return: A FileAPIDownload object.
        """
        response = await self._api.get_file_by_path(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_path=path,
            version_id=version_id,
        )
        metadata = self._metadata_from_endpoint_model(response)
        return FileAPIDownload(connector=self._connector, metadata=metadata, initial_url=response.download)

    async def prepare_download_by_id(self, file_id: UUID, version_id: str | None = None) -> FileAPIDownload:
        """Prepares a file for download by ID.

        :param file_id: UUID of the file.
        :param version_id: Version of the file.

        :return: A FileAPIDownload object.
        """
        response = await self._api.get_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(file_id),
            version_id=version_id,
        )
        metadata = self._metadata_from_endpoint_model(response)
        return FileAPIDownload(connector=self._connector, metadata=metadata, initial_url=response.download)

    async def prepare_upload_by_path(self, path: str) -> FileAPIUpload:
        """Prepares a file for upload by path. If the file already exists, a new version will be created.

        :param path: Path the file is being uploaded to.

        :return: A FileAPIUpload object.
        """
        response = await self._api.upsert_file_by_path(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_path=path,
        )
        return FileAPIUpload(
            connector=self._connector,
            environment=self._environment,
            file_id=response.file_id,
            version_id=response.version_id,
            initial_url=response.upload,
        )

    async def prepare_upload_by_id(self, file_id: UUID) -> FileAPIUpload:
        """Prepares a file for upload by ID. The file_id must be the ID of an existing file, for which a new version
        will be created.

        :param file_id: UUID of the file.

        :return: A FileAPIUpload object.
        """
        response = await self._api.update_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(file_id),
        )
        return FileAPIUpload(
            connector=self._connector,
            environment=self._environment,
            file_id=file_id,
            version_id=response.version_id,
            initial_url=response.upload,
        )

    async def delete_file_by_path(self, path: str) -> None:
        """Deletes a file by path.

        :param path: Path of the file to delete.
        """
        await self._api.delete_file_by_path(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_path=path,
        )

    async def delete_file_by_id(self, file_id: UUID) -> None:
        """Deletes a file by ID.

        :param file_id: UUID of the file to delete.
        """
        await self._api.delete_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(file_id),
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
from pathlib import Path
from uuid import UUID

from evo.common import APIConnector, Environment, ICache
from evo.common.io import Download, Upload

from .data import FileMetadata
from .endpoints import FileV2Api

__all__ = [
    "FileAPIDownload",
    "FileAPIUpload",
]

_CACHE_SCOPE = "filev2"
_INVALID_CHARS = {"<", ">", ":", '"', "/", "\\", "|", "?", "*"} | {chr(i) for i in range(32)}
_INVALID_TRAILING_CHARS = {" ", "."}
_INVALID_NAMES = {
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
}


def _quote_char(c: str) -> str:
    return "%{:02X}".format(ord(c))


def _make_name_safe(name: str) -> str:
    # Replace invalid characters with URL encoded form.
    new_name = "".join(_quote_char(c) if c in _INVALID_CHARS else c for c in name)
    # Replace invalid trailing characters with URL encoded form.
    if new_name and new_name[-1] in _INVALID_TRAILING_CHARS:
        new_name = new_name[:-1] + _quote_char(new_name[-1])
    # Fix reserved names by replacing 'aux' with 'aux_' and 'aux.txt' with 'aux_.txt'.
    parts = new_name.split(".", 1)
    if parts[0].upper() in _INVALID_NAMES:
        if len(parts) == 1:
            new_name += "_"
        else:
            new_name = f"{parts[0]}_.{parts[1]}"
    return new_name


class _FileIOMixin:
    def __init__(self, connector: APIConnector, initial_url: str) -> None:
        self._api = FileV2Api(connector)
        self._mutex = asyncio.Lock()
        self._initial_url = initial_url

    async def _get_initial_url(self) -> str:
        async with self._mutex:
            if (url := self._initial_url) is not None:
                self._initial_url = None
                return url


class FileAPIUpload(Upload, _FileIOMixin):
    """A context for uploading a file to the File API.

    Do not use this class directly. Instead, use the `FileAPIClient.prepare_upload_by_*` methods.
    """

    def __init__(
        self, connector: APIConnector, environment: Environment, file_id: UUID, version_id: str, initial_url: str
    ) -> None:
        """
        :param connector: The connector to use for the API calls.
        :param environment: The environment the file is being uploaded to.
        :param file_id: The ID of the file that will be uploaded.
        :param version_id: The version ID of the file that will be uploaded.
        :param initial_url: The initial URL to use for the upload.
        """
        super().__init__(connector, initial_url)
        self._environment = environment
        self._id = file_id
        self._version_id = version_id

    @property
    def file_id(self) -> UUID:
        """The ID of the file that is being uploaded."""
        return self._id

    @property
    def version_id(self) -> str:
        """The ID of the version that will be created after the upload completes."""
        return self._version_id

    @property
    def label(self) -> str:
        """The file and version ID of the file that is being uploaded."""
        return f"{self._id}?version_id={self._version_id}"

    async def get_upload_url(self) -> str:
        """Generate a URL that will be used to upload the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The upload URL.
        """
        # Use the initial URL first.
        if (url := await self._get_initial_url()) is not None:
            return url

        # Otherwise, generate a new URL.
        response = await self._api.update_file_by_id(
            organisation_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            file_id=str(self._id),
            version_id=self._version_id,
        )
        return response.upload


class FileAPIDownload(Download[FileMetadata], _FileIOMixin):
    """A context for downloading a file from the File API.

    Do not use this class directly. Instead, use the `FileAPIClient.prepare_download_by_*` methods.
    """

    def __init__(self, connector: APIConnector, metadata: FileMetadata, initial_url: str) -> None:
        """
        :param connector: The connector to use for the API calls.
        :param metadata: The metadata of the file that will be downloaded.
        :param initial_url: The initial URL to use for the download.
        """
        super().__init__(connector, initial_url)
        self._metadata = metadata

    @property
    def label(self) -> str:
        """The file and version ID for the file that is being downloaded."""
        return f"{self.metadata.id}?version_id={self.metadata.version_id}"

    @property
    def metadata(self) -> FileMetadata:
        """The metadata of the file that is being downloaded."""
        return self._metadata

    def _get_cache_location(self, cache: ICache) -> Path:
        download_dir = (
            cache.get_location(self.metadata.environment, _CACHE_SCOPE)
            / str(self.metadata.id)
            / self.metadata.version_id
        )
        download_dir.mkdir(parents=True, exist_ok=True)
        return download_dir / _make_name_safe(self.metadata.name)

    async def get_download_url(self) -> str:
        """Generate a URL that will be used to download the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The download URL.
        """
        # Use the initial URL first.
        if (url := await self._get_initial_url()) is not None:
            return url

        # Otherwise, generate a new URL.
        response = await self._api.get_file_by_id(
            organisation_id=str(self.metadata.environment.org_id),
            workspace_id=str(self.metadata.environment.workspace_id),
            file_id=str(self.metadata.id),
            version_id=self.metadata.version_id,
        )
        return response.download



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Protocol

# `evo-blockmodels` uses protocols for annotating some pyarrow types, because:
# - pyarrow is optional, but type annotations are not.
# - pyarrow has poor type checker support.
#
# These protocols should be treated as aliases for the corresponding pyarrow types.
# Any required interfaces from the corresponding pyarrow types should be added to these protocols as needed.


class DataType(Protocol):
    """Pyarrow data type.

    https://arrow.apache.org/docs/python/generated/pyarrow.DataType.html
    """

    ...


class Schema(Protocol):
    """Pyarrow schema.

    https://arrow.apache.org/docs/python/generated/pyarrow.Schema.html
    """

    @property
    def names(self) -> list[str]:
        """The schema's field names."""
        ...

    @property
    def types(self) -> list[DataType]:
        """The schema's field types."""
        ...


class Table(Protocol):
    """Pyarrow table.

    https://arrow.apache.org/docs/python/generated/pyarrow.Table.html
    """

    @property
    def schema(self) -> Schema:
        """Schema of the table and its columns."""
        ...

    @property
    def num_columns(self) -> int:
        """Number of columns in this table."""
        ...

    @property
    def num_rows(self) -> int:
        """Number of rows in this table.

        Due to the definition of a table, all columns have the same number of rows.
        """
        ...

    def to_pandas(self) -> DataFrame:
        """Convert to a pandas-compatible NumPy array or DataFrame, as appropriate"""


class DataFrame(Protocol):
    """Pandas DataFrame.

    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
    """



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


try:
    # Import the table type for backwards compatibility. This should be removed in a future release.
    from pyarrow import Table  # noqa: F401
except ImportError:
    raise ImportError("pyarrow is required to use the utils package in evo-objects")

try:
    # Import the dataframe type for backwards compatibility. This should be removed in a future release.
    from pandas import DataFrame  # noqa: F401
except ImportError:
    DataFrame = None  # type: ignore

from .data import ObjectDataClient
from .table_formats import all_known_formats, get_known_format
from .tables import ArrowTableFormat, BaseTableFormat, KnownTableFormat

# We _used_ to export Table and DataFrame from this package as custom protocols, but we are using the actual
# pyarrow.Table and pandas.DataFrame types now. We are importing these types here from pyarrow and pandas
# for backwards compatibility, but they are no longer explicitly exported as the exports should be
# removed in a future release.

__all__ = [
    "ArrowTableFormat",
    "BaseTableFormat",
    "KnownTableFormat",
    "ObjectDataClient",
    "all_known_formats",
    "get_known_format",
]
